# -*- coding: utf-8 -*-


import pytest

import aerospike

from aerospike_helpers import expressions as exp
from aerospike_helpers.operations import operations as op
from aerospike import exception as e
from .test_base_class import TestBaseClass
from .as_status_codes import AerospikeStatus


class TestBatchOperate(TestBaseClass):
    @pytest.fixture(autouse=True)
    def setup(self, request, connection_with_config_funcs):
        as_connection = connection_with_config_funcs

        if self.server_version < [6, 0]:
            pytest.mark.xfail(reason="Servers older than 6.0 do not support batch operate.")
            pytest.xfail()

        self.test_ns = "test"
        self.test_set = "demo"
        self.keys = []
        self.batch_size = 5

        for i in range(self.batch_size):
            key = ("test", "demo", i)
            rec = {
                "count": i,
                "name": "name10",
                "t": True,
                "age": 10,
                "balance": 100,
                "key": 10,
                "ilist_bin": [
                    i,
                    1,
                    2,
                    6,
                ],
                "imap_bin": {
                    1: 1,
                    2: 2,
                    3: 6,
                },
            }
            as_connection.put(key, rec)
            self.keys.append(key)

        def teardown():
            for i in range(self.batch_size):
                key = ("test", "demo", i)
                as_connection.remove(key)

        request.addfinalizer(teardown)

    @pytest.mark.parametrize(
        "name, keys, ops, policy_batch, policy_batch_write, exp_res, exp_rec",
        [
            (
                "simple-write",
                [("test", "demo", 0)],
                [op.write("count", 2), op.read("count")],
                None,
                None,
                [AerospikeStatus.AEROSPIKE_OK],
                [{"count": 2}],
            ),
            (
                "simple-write2",
                [("test", "demo", 1)],
                [op.write("count", 3), op.read("count")],
                {},
                {},
                [AerospikeStatus.AEROSPIKE_OK],
                [{"count": 3}],
            ),
            (
                "simple-write-policy-batch",
                [("test", "demo", 0)],
                [op.write("count", 7), op.read("count")],
                {
                    "total_timeout": 2000,
                    "max_retries": 2,
                    "allow_inline_ssd": True,
                    "respond_all_keys": False,
                    "expressions": exp.Eq(
                        exp.ListGetByRank(
                            None, aerospike.LIST_RETURN_VALUE, exp.ResultType.INTEGER, 0, exp.ListBin("ilist_bin")
                        ),
                        0,
                    ).compile(),
                },
                {},
                [AerospikeStatus.AEROSPIKE_OK],
                [{"count": 7}],
            ),
            (
                "simple-write-policy-batch-write",
                [("test", "demo", 0)],
                [op.write("count", 7), op.read("count")],
                {},
                {
                    "key": aerospike.POLICY_KEY_SEND,
                    "commit_level": aerospike.POLICY_COMMIT_LEVEL_MASTER,
                    "gen": aerospike.POLICY_GEN_IGNORE,
                    "exists": aerospike.POLICY_EXISTS_UPDATE,
                    "durable_delete": False,
                    "expressions": exp.Eq(exp.IntBin("count"), 0).compile(),
                },
                [AerospikeStatus.AEROSPIKE_OK],
                [{"count": 7}],
            ),
            (
                "simple-write-policy-both",
                [("test", "demo", 0)],
                [op.write("count", 7), op.read("count")],
                {
                    "total_timeout": 2000,
                    "max_retries": 2,
                    "allow_inline_ssd": True,
                    "respond_all_keys": False,
                    "expressions": exp.Eq(
                        exp.ListGetByRank(
                            None, aerospike.LIST_RETURN_VALUE, exp.ResultType.INTEGER, 0, exp.ListBin("ilist_bin")
                        ),
                        1,
                    ).compile(),
                },
                {
                    "key": aerospike.POLICY_KEY_SEND,
                    "commit_level": aerospike.POLICY_COMMIT_LEVEL_MASTER,
                    "gen": aerospike.POLICY_GEN_IGNORE,
                    "exists": aerospike.POLICY_EXISTS_UPDATE,
                    "durable_delete": False,
                    "expressions": exp.Eq(exp.IntBin("count"), 0).compile(),  # this expression takes precedence
                },
                [AerospikeStatus.AEROSPIKE_OK],
                [{"count": 7}],
            ),
        ],
    )
    def test_batch_operate_pos(self, name, keys, ops, policy_batch, policy_batch_write, exp_res, exp_rec):
        """
        Test batch_operate positive.
        """

        res = self.as_connection.batch_operate(keys, ops, policy_batch, policy_batch_write)

        for i, batch_rec in enumerate(res.batch_records):
            assert batch_rec.result == exp_res[i]
            assert batch_rec.record[2] == exp_rec[i]
            assert batch_rec.key[:3] == keys[i]  # checking key
            assert batch_rec.record[0][:3] == keys[i]  # checking key in record

    def test_batch_operate_many_pos(self):
        """
        Test batch operate with many keys.
        """

        keys = [("test", "demo", i) for i in range(100, 1000)]

        ops = [
            op.write("count", 10),
            op.read("count"),
        ]

        policy_batch = {}

        policy_batch_write = {}

        try:
            for key in keys:
                self.as_connection.put(key, {"count": 0})

            res = self.as_connection.batch_operate(keys, ops, policy_batch, policy_batch_write)

            for i, batch_rec in enumerate(res.batch_records):
                assert batch_rec.result == AerospikeStatus.AEROSPIKE_OK
                assert batch_rec.record[2] == {"count": 10}
                assert batch_rec.key[:3] == keys[i]  # checking key
                assert batch_rec.record[0][:3] == keys[i]  # checking key in record
        except Exception as ex:
            raise (ex)
        finally:
            for key in keys:
                self.as_connection.remove(key)

    @pytest.mark.parametrize(
        "name, keys, ops, policy_batch, policy_batch_write, exp_res",
        [
            (
                "bad-key",
                [("bad-key", i) for i in range(1000)],
                [op.write("count", 2), op.read("count")],
                {},
                {},
                e.ParamError,
            ),
            (
                "bad-ops",
                [("test", "demo", 1)],
                [op.write("count", 2), op.read("count"), ["bad-op"]],
                {
                    "respond_all_keys": False,
                },
                {},
                e.ParamError,
            ),
            (
                "bad-ops2",
                [("test", "demo", 1)],
                {"bad": "ops"},
                {
                    "respond_all_keys": False,
                },
                {},
                e.ParamError,
            ),
            (
                "bad-batch-policy",
                [("test", "demo", 1)],
                [
                    op.write("count", 2),
                    op.read("count"),
                ],
                ["bad-batch-policy"],
                {},
                e.ParamError,
            ),
            (
                "bad-batch-write-policy",
                [("test", "demo", 1)],
                [
                    op.write("count", 2),
                    op.read("count"),
                ],
                {},
                ["bad-batch-write-policy"],
                e.ParamError,
            ),
        ],
    )
    def test_batch_operate_neg(self, name, keys, ops, policy_batch, policy_batch_write, exp_res):
        """
        Test batch_operate negative.
        """

        with pytest.raises(exp_res):
            self.as_connection.batch_operate(keys, ops, policy_batch, policy_batch_write)

    def test_batch_operate_neg_connection(self):
        """
        Test batch_operate negative with bad connection.
        """

        keys = []
        ops = ["doesn't_matter"]

        exp_res = e.ClientError

        with pytest.raises(exp_res):
            bad_client = aerospike.client({"hosts": [("bad_addr", 3000)]})
            bad_client.batch_operate(keys, ops)



# -*- coding: utf-8 -*-
################################################################################
# Copyright 2013-2021 Aerospike, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

from __future__ import print_function

import aerospike
import sys

from optparse import OptionParser

################################################################
# Option Parsing
################################################################

usage = "usage: %prog [options] key"

optparser = OptionParser(usage=usage, add_help_option=False)

optparser.add_option(
  "--help", dest="help", action="store_true",
  help="Displays this message.")

optparser.add_option(
  "-h", "--host", dest="host", type="string", default="127.0.0.1", metavar="<ADDRESS>",
  help="Address of Aerospike server.")

optparser.add_option(
  "-p", "--port", dest="port", type="int", default=3000, metavar="<PORT>",
  help="Port of the Aerospike server.")

optparser.add_option(
  "-n", "--namespace", dest="namespace", type="string", default="test", metavar="<NS>",
  help="Port of the Aerospike server.")

optparser.add_option(
  "-s", "--set", dest="set", type="string", default="demo", metavar="<SET>",
  help="Port of the Aerospike server.")

(options, args) = optparser.parse_args()

if options.help:
  optparser.print_help()
  print()
  sys.exit(1)

if len(args) != 1:
  optparser.print_help()
  print()
  sys.exit(1)

################################################################
# Connect to Cluster
################################################################

config = {
  'hosts': [ (options.host, options.port) ]
}

client = aerospike.client(config).connect()

################################################################
# Perform Operation
################################################################

rc = 0
key = args.pop()

try:

  (key, metadata) = client.key(options.namespace, options.set, key).exists()

  if metadata != None:
    print(metadata)
    print("---")
    print("OK, 1 record found.")
  else:
    print('error: Not Found.', file=sys.stderr)
    rc = 1

except Exception as e:
  print("error: {0}".format(e), file=sys.stderr)
  rc = 1

################################################################
# Close Connection to Cluster
################################################################

client.close()

################################################################
# Exit
################################################################

sys.exit(rc)



# -*- coding: utf-8 -*-
import pytest
import time

import aerospike
from aerospike_helpers import expressions as exp
from aerospike_helpers.operations import operations
from aerospike import exception, predicates

TEST_NS = "test"
TEST_SET = "background_q_e"
TEST_UDF_MODULE = "query_apply"
TEST_UDF_FUNCTION = "mark_as_applied"
# Hack to get long to exist in Python 3
try:
    long
except NameError:
    long = int


def wait_for_job_completion(as_connection, job_id):
    """
    Blocks until the job has completed
    """
    while True:
        response = as_connection.job_info(job_id, aerospike.JOB_QUERY)
        if response["status"] != aerospike.JOB_STATUS_INPROGRESS:
            break
        time.sleep(0.1)


def add_indexes_to_client(client):
    try:
        client.index_integer_create(TEST_NS, TEST_SET, "number", "test_background_number_idx")
    except exception.IndexFoundError:
        pass


def add_test_udf(client):
    policy = {}
    client.udf_put("query_apply.lua", 0, policy)


def drop_test_udf(client):
    client.udf_remove("query_apply.lua")


def remove_indexes_from_client(client):
    client.index_remove(TEST_NS, "test_background_number_idx")


def validate_records(client, keys, validator):
    for key in keys:
        _, _, rec = client.get(key)
        assert validator(rec)


# Add records around the test
@pytest.fixture(scope="function")
def clean_test_background(as_connection):
    keys = [(TEST_NS, TEST_SET, i) for i in range(500)]
    for i, key in enumerate(keys):
        as_connection.put(key, {"number": i})
    yield
    for i, key in enumerate(keys):
        as_connection.remove(key)


class TestQueryApply(object):

    # These functions will run once for this test class, and do all of the
    # required setup and teardown
    connection_setup_functions = (add_test_udf, add_indexes_to_client)
    connection_teardown_functions = (drop_test_udf, remove_indexes_from_client)

    @pytest.fixture(autouse=True)
    def setup(self, connection_with_config_funcs):
        pass

    def test_background_execute_return_val(self, clean_test_background):
        """
        Ensure that Query.execute_background() returns an int like object
        """
        test_bin = "tz"
        query = self.as_connection.query(TEST_NS, TEST_SET)
        query.apply(TEST_UDF_MODULE, TEST_UDF_FUNCTION, [test_bin])
        res = query.execute_background()
        assert isinstance(res, (int, long))

    def test_background_execute_return_val_with_ttl(self, clean_test_background):
        """
        Ensure that Query.execute_background() returns an int like object
        """
        test_bin = "tz"
        query = self.as_connection.query(TEST_NS, TEST_SET)
        query.max_records = 100
        query.ttl = aerospike.TTL_DONT_UPDATE
        query.apply(TEST_UDF_MODULE, TEST_UDF_FUNCTION, [test_bin])
        res = query.execute_background()
        assert isinstance(res, (int, long))

    def test_background_execute_no_predicate(self, clean_test_background):
        """
        Ensure that Query.execute_background() gets applied to all records
        """
        test_bin = "t2222"
        keys = [(TEST_NS, TEST_SET, i) for i in range(500)]

        query = self.as_connection.query(TEST_NS, TEST_SET)
        query.apply(TEST_UDF_MODULE, TEST_UDF_FUNCTION, [test_bin])
        query.execute_background()
        # Give time for the query to finish

        time.sleep(5)
        # wait_for_job_completion(self.as_connection, job_id)

        validate_records(self.as_connection, keys, lambda rec: rec[test_bin] == "aerospike")

    def test_background_execute_exp_everywhere(self, clean_test_background):
        """
        Ensure that Query.execute_background() gets applied to records that match the exp
        """
        from .test_base_class import TestBaseClass

        if TestBaseClass.major_ver >= 6 or (TestBaseClass.major_ver >= 5 and TestBaseClass.minor_ver >= 7):
            # print("TestBaseClass.major_ver:", TestBaseClass.major_ver, "TestBaseClass.minor_ver:", TestBaseClass.minor_ver)  # noqa: E501
            pytest.skip("It deprecated and it only applies to < 5.7 earlier and enterprise edition")

        test_bin = "tpred"
        keys = [(TEST_NS, TEST_SET, i) for i in range(500)]

        expr = exp.Or(exp.Eq(exp.IntBin("number"), 2), exp.Eq(exp.IntBin("number"), 3))

        # number_predicate = predicates.equals('number', 3)

        policy = {"expressions": expr.compile()}

        query = self.as_connection.query(TEST_NS, TEST_SET)
        # query.where(number_predicate)
        query.apply(TEST_UDF_MODULE, TEST_UDF_FUNCTION, [test_bin])
        query.execute_background(policy)
        # Give time for the query to finish
        time.sleep(5)
        # wait_for_job_completion(self.as_connection, job_id)

        for key in keys:
            _, _, bins = self.as_connection.get(key)
            if bins["number"] == 2 or bins["number"] == 3:
                assert bins[test_bin] == "aerospike"
            else:
                assert bins.get(test_bin) is None

    @pytest.mark.xfail(reason="predicate and exp used at same time")
    def test_background_execute_exp_and_predicate(self, clean_test_background):
        """
        Ensure that Query.execute_background() gets applied to records that match the predicate
        NOTE: the predicate overrides the exp
        """
        test_bin = "tpredold"
        keys = [(TEST_NS, TEST_SET, i) for i in range(500)]

        expr = exp.Or(exp.Eq(exp.IntBin("number"), 2), exp.Eq(exp.IntBin("number"), 3))

        number_predicate = predicates.equals("number", 4)

        policy = {"expressions": expr.compile()}

        query = self.as_connection.query(TEST_NS, TEST_SET)
        query.where(number_predicate)
        query.apply(TEST_UDF_MODULE, TEST_UDF_FUNCTION, [test_bin])
        query.execute_background(policy)
        # Give time for the query to finish
        time.sleep(5)

        for key in keys:
            _, _, bins = self.as_connection.get(key)
            if bins["number"] == 4:
                assert bins[test_bin] == "aerospike"
            else:
                assert bins.get(test_bin) is None

    def test_background_execute_with_ops_and_exp(self, clean_test_background):
        """
        Ensure that Query.execute_background() applies ops to records that match the expressions.
        """
        from .test_base_class import TestBaseClass

        if TestBaseClass.major_ver >= 6 or (TestBaseClass.major_ver >= 5 and TestBaseClass.minor_ver >= 7):
            # print("TestBaseClass.major_ver:", TestBaseClass.major_ver, "TestBaseClass.minor_ver:", TestBaseClass.minor_ver)  # noqa: E501
            pytest.skip("It deprecated and it only applies to < 5.7 earlier and enterprise edition")

        test_bin = "tops_preds"
        keys = [(TEST_NS, TEST_SET, i) for i in range(500)]

        query = self.as_connection.query(TEST_NS, TEST_SET)

        ops = [operations.write(test_bin, "new_val")]

        expr = exp.Or(exp.Eq(exp.IntBin("number"), 2), exp.Eq(exp.IntBin("number"), 3))

        policy = {"expressions": expr.compile()}

        query.add_ops(ops)
        query.execute_background(policy=policy)
        # Give time for the query to finish
        time.sleep(5)

        for key in keys:
            _, _, bins = self.as_connection.get(key)
            if bins["number"] == 2 or bins["number"] == 3:
                assert bins[test_bin] == "new_val"
            else:
                assert bins.get(test_bin) is None

    def test_background_execute_with_ops(self, clean_test_background):
        """
        Ensure that Query.execute_background() applies ops to all records
        """
        test_bin = "tops"
        keys = [(TEST_NS, TEST_SET, i) for i in range(500)]

        query = self.as_connection.query(TEST_NS, TEST_SET)
        # query.apply(TEST_UDF_MODULE, TEST_UDF_FUNCTION, [test_bin])

        ops = [operations.write(test_bin, "new_val")]

        query.add_ops(ops)
        query.execute_background()
        # Give time for the query to finish
        time.sleep(5)

        validate_records(self.as_connection, keys, lambda rec: rec[test_bin] == "new_val")

    def test_background_execute_with_ops_and_preds(self, clean_test_background):
        """
        Ensure that Query.execute_background() applies ops to records that match the predicate
        """
        test_bin = "t1"
        keys = [(TEST_NS, TEST_SET, i) for i in range(500)]

        query = self.as_connection.query(TEST_NS, TEST_SET)
        number_predicate = predicates.equals("number", 3)
        # query.apply(TEST_UDF_MODULE, TEST_UDF_FUNCTION, [test_bin])

        ops = [operations.append(test_bin, "new_val")]

        query.add_ops(ops)
        query.where(number_predicate)
        query.execute_background()
        # Give time for the query to finish
        time.sleep(5)

        _, _, num_5_record = self.as_connection.get((TEST_NS, TEST_SET, 5))
        assert num_5_record.get(test_bin) is None

        _, _, num_3_record = self.as_connection.get((TEST_NS, TEST_SET, 3))
        assert num_3_record[test_bin] == "new_val"

        # cleanup
        ops = [operations.write(test_bin, "aerospike")]

        query = self.as_connection.query(TEST_NS, TEST_SET)
        query.add_ops(ops)
        query.execute_background()
        time.sleep(3)

        validate_records(self.as_connection, keys, lambda rec: rec[test_bin] == "aerospike")

    def test_background_execute_sindex_predicate(self, clean_test_background):
        """
        Ensure that Query.execute_background() only applies to records matched by
        the specified predicate
        """
        test_bin = "t3"
        keys = [(TEST_NS, TEST_SET, i) for i in range(500)]
        number_predicate = predicates.equals("number", 5)

        query = self.as_connection.query(TEST_NS, TEST_SET)
        query.where(number_predicate)
        query.apply(TEST_UDF_MODULE, TEST_UDF_FUNCTION, [test_bin])
        query.execute_background()
        # Give time for the query to finish
        time.sleep(5)
        keys = [(TEST_NS, TEST_SET, i) for i in range(500) if i != 5]
        validate_records(self.as_connection, keys, lambda rec: test_bin not in rec)
        _, _, num_5_record = self.as_connection.get((TEST_NS, TEST_SET, 5))
        assert num_5_record[test_bin] == "aerospike"

    def test_background_execute_sindex_exp(self, clean_test_background):
        """
        Ensure that Query.execute_background() only applies to records matched by
        the specified predicate
        """
        from .test_base_class import TestBaseClass

        if TestBaseClass.major_ver >= 6 or (TestBaseClass.major_ver >= 5 and TestBaseClass.minor_ver >= 7):
            # print("TestBaseClass.major_ver:", TestBaseClass.major_ver, "TestBaseClass.minor_ver:", TestBaseClass.minor_ver)  # noqa: E501
            pytest.skip("It deprecated and it only applies to < 5.7 earlier and enterprise edition")

        test_bin = "t4"
        keys = [(TEST_NS, TEST_SET, i) for i in range(500)]

        #  rec['number'] < 10
        expr = exp.LT(exp.IntBin("number"), 10)
        policy = {"expressions": expr.compile()}

        query = self.as_connection.query(TEST_NS, TEST_SET)
        query.apply(TEST_UDF_MODULE, TEST_UDF_FUNCTION, [test_bin])
        query.execute_background(policy=policy)
        # Give time for the query to finish
        time.sleep(5)

        # Records with number > 10 should not have had the UDF applied
        validate_records(self.as_connection, keys[10:], lambda rec: test_bin not in rec)
        #  Records with number < 10 should have had the udf applied
        validate_records(self.as_connection, keys[:10], lambda rec: rec[test_bin] == "aerospike")

    def test_background_execute_with_policy(self, clean_test_background):
        """
        Ensure that Query.execute_background() returns an int like object
        """
        test_bin = "t5"
        query = self.as_connection.query(TEST_NS, TEST_SET)
        query.apply(TEST_UDF_MODULE, TEST_UDF_FUNCTION, [test_bin])
        res = query.execute_background({"socket_timeout": 10000})
        assert isinstance(res, (int, long))

    def test_background_execute_with_policy_kwarg(self, clean_test_background):
        """
        Ensure that Query.execute_background() returns an int like object
        """
        test_bin = "t6"
        query = self.as_connection.query(TEST_NS, TEST_SET)
        query.apply(TEST_UDF_MODULE, TEST_UDF_FUNCTION, [test_bin])
        res = query.execute_background(policy={})
        assert isinstance(res, (int, long))

    def test_background_execute_with_invalid_policy_type(self, clean_test_background):
        """
        Ensure that Query.execute_background() returns an int like object
        """
        test_bin = "t6"
        query = self.as_connection.query(TEST_NS, TEST_SET)
        query.apply(TEST_UDF_MODULE, TEST_UDF_FUNCTION, [test_bin])
        # Policy needs to be a dict. Not a string
        with pytest.raises(exception.ParamError):
            query.execute_background("Honesty is the best Policy")



import pytest
from aerospike import exception as e
from .as_errors import (
    AEROSPIKE_ERR_ASYNC_CONNECTION,
    AEROSPIKE_ERR_BATCH_DISABLED,
    AEROSPIKE_ERR_BATCH_MAX_REQUESTS_EXCEEDED,
    AEROSPIKE_ERR_BATCH_QUEUES_FULL,
    AEROSPIKE_ERR_CLIENT_ABORT,
    AEROSPIKE_ERR_FAIL_ELEMENT_EXISTS,
    AEROSPIKE_ERR_FAIL_ELEMENT_NOT_FOUND,
    AEROSPIKE_ERR_INVALID_NODE,
    AEROSPIKE_ERR_NO_MORE_CONNECTIONS,
    AEROSPIKE_ERR_QUERY_ABORTED,
    AEROSPIKE_ERR_SCAN_ABORTED,
    AEROSPIKE_ERR_TLS_ERROR,
)


@pytest.mark.parametrize(
    "error, error_name, error_code, base",
    (
        (e.TLSError, "TLSError", AEROSPIKE_ERR_TLS_ERROR, e.ClientError),
        (e.InvalidNodeError, "InvalidNodeError", AEROSPIKE_ERR_INVALID_NODE, e.ClientError),
        (e.NoMoreConnectionsError, "NoMoreConnectionsError", AEROSPIKE_ERR_NO_MORE_CONNECTIONS, e.ClientError),
        (e.AsyncConnectionError, "AsyncConnectionError", AEROSPIKE_ERR_ASYNC_CONNECTION, e.ClientError),
        (e.ClientAbortError, "ClientAbortError", AEROSPIKE_ERR_CLIENT_ABORT, e.ClientError),
        (e.ScanAbortedError, "ScanAbortedError", AEROSPIKE_ERR_SCAN_ABORTED, e.ServerError),
        (e.ElementNotFoundError, "ElementNotFoundError", AEROSPIKE_ERR_FAIL_ELEMENT_NOT_FOUND, e.ServerError),
        (e.ElementExistsError, "ElementExistsError", AEROSPIKE_ERR_FAIL_ELEMENT_EXISTS, e.ServerError),
        (e.BatchDisabledError, "BatchDisabledError", AEROSPIKE_ERR_BATCH_DISABLED, e.ServerError),
        (e.BatchMaxRequestError, "BatchMaxRequestError", AEROSPIKE_ERR_BATCH_MAX_REQUESTS_EXCEEDED, e.ServerError),
        (e.BatchQueueFullError, "BatchQueueFullError", AEROSPIKE_ERR_BATCH_QUEUES_FULL, e.ServerError),
        (e.QueryAbortedError, "QueryAbortedError", AEROSPIKE_ERR_QUERY_ABORTED, e.ServerError),
    ),
)
def test_error_codes(error, error_name, error_code, base):
    with pytest.raises(error) as test_error:
        raise error

    test_error = test_error.value

    assert test_error.code == error_code
    assert type(test_error).__name__ == error_name
    assert issubclass(type(test_error), base)



# -*- coding: utf-8 -*-
import pytest
from .test_base_class import TestBaseClass

import aerospike
from aerospike import exception as e

# OPERATIONS
# aerospike.OPERATOR_WRITE
# aerospike.OPERATOR_APPEND
# aerospike.OPERATOR_PREPEND
# aerospike.OPERATOR_INCR
# aerospike.OPERATOR_READ
# aerospike.OPERATOR_TOUCH
# aerospike.OP_LIST_APPEND
# aerospike.OP_LIST_APPEND_ITEMS
# aerospike.OP_LIST_INSERT
# aerospike.OP_LIST_INSERT_ITEMS
# aerospike.OP_LIST_POP
# aerospike.OP_LIST_POP_RANGE
# aerospike.OP_LIST_REMOVE
# aerospike.OP_LIST_REMOVE_RANGE
# aerospike.OP_LIST_CLEAR
# aerospike.OP_LIST_SET
# aerospike.OP_LIST_GET
# aerospike.OP_LIST_GET_RANGE
# aerospike.OP_LIST_TRIM
# aerospike.OP_LIST_SIZE
# aerospike.OP_MAP_SET_POLICY
# aerospike.OP_MAP_PUT
# aerospike.OP_MAP_PUT_ITEMS
# aerospike.OP_MAP_INCREMENT
# aerospike.OP_MAP_DECREMENT
# aerospike.OP_MAP_SIZE
# aerospike.OP_MAP_CLEAR
# aerospike.OP_MAP_REMOVE_BY_KEY
# aerospike.OP_MAP_REMOVE_BY_KEY_LIST
# aerospike.OP_MAP_REMOVE_BY_KEY_RANGE
# aerospike.OP_MAP_REMOVE_BY_VALUE
# aerospike.OP_MAP_REMOVE_BY_VALUE_LIST
# aerospike.OP_MAP_REMOVE_BY_VALUE_RANGE
# aerospike.OP_MAP_REMOVE_BY_INDEX
# aerospike.OP_MAP_REMOVE_BY_INDEX_RANGE
# aerospike.OP_MAP_REMOVE_BY_RANK
# aerospike.OP_MAP_REMOVE_BY_RANK_RANGE
# aerospike.OP_MAP_GET_BY_KEY
# aerospike.OP_MAP_GET_BY_KEY_RANGE
# aerospike.OP_MAP_GET_BY_VALUE
# aerospike.OP_MAP_GET_BY_VALUE_RANGE
# aerospike.OP_MAP_GET_BY_INDEX
# aerospike.OP_MAP_GET_BY_INDEX_RANGE
# aerospike.OP_MAP_GET_BY_RANK
# aerospike.OP_MAP_GET_BY_RANK_RANGE


class TestOperate(object):
    def setup_class(cls):
        """
        Setup class.
        """
        cls.client_no_typechecks = TestBaseClass.get_new_connection({"strict_types": False})

    def teardown_class(cls):
        TestOperate.client_no_typechecks.close()

    @pytest.fixture(autouse=True)
    def setup(self, request, as_connection):
        """
        Setup Method
        """
        keys = []

        for i in range(5):
            key = ("test", "demo", i)
            rec = {"name": "name%s" % (str(i)), "age": i}
            as_connection.put(key, rec)

        key = ("test", "demo", 6)
        rec = {"age": 6.3}
        as_connection.put(key, rec)
        keys.append(key)

        key = ("test", "demo", "bytearray_key")
        rec = {"bytearray_bin": bytearray("asd;as[d'as;d", "utf-8")}
        as_connection.put(key, rec)

        keys.append(key)

        key = ("test", "demo", "list_key")
        rec = {"int_bin": [1, 2, 3, 4], "string_bin": ["a", "b", "c", "d"]}
        as_connection.put(key, rec)

        keys.append(key)

        key = ("test", "demo", "existing_key")
        rec = {"dict": {"a": 1}, "bytearray": bytearray("abc", "utf-8"), "float": 3.4, "list": ["a"]}
        as_connection.put(key, rec)

        keys.append(key)

        def teardown():
            """
            Teardown Method
            """
            for i in range(5):
                key = ("test", "demo", i)
                try:
                    as_connection.remove(key)
                except e.RecordNotFound:
                    pass
            for key in keys:
                try:
                    as_connection.remove(key)
                except e.RecordNotFound:
                    pass

        request.addfinalizer(teardown)

    @pytest.mark.parametrize(
        "key, llist, expected",
        [
            (
                ("test", "demo", 1),
                [
                    {"op": aerospike.OPERATOR_PREPEND, "bin": "name", "val": "ram"},
                    {"op": aerospike.OPERATOR_INCR, "bin": "age", "val": 3},
                    {"op": aerospike.OPERATOR_READ, "bin": "name"},
                ],
                {"name": "ramname1"},
            ),
            (
                ("test", "demo", 1),  # with_write_float_value
                [
                    {"op": aerospike.OPERATOR_WRITE, "bin": "write_bin", "val": {"no": 89.8}},
                    {"op": aerospike.OPERATOR_READ, "bin": "write_bin"},
                ],
                {"write_bin": {"no": 89.8}},
            ),
            (
                ("test", "demo", 1),  # write positive
                [
                    {"op": aerospike.OPERATOR_WRITE, "bin": "write_bin", "val": {"no": 89}},
                    {"op": aerospike.OPERATOR_READ, "bin": "write_bin"},
                ],
                {"write_bin": {"no": 89}},
            ),
            (
                ("test", "demo", 1),  # write_tuple_positive
                [
                    {"op": aerospike.OPERATOR_WRITE, "bin": "write_bin", "val": tuple("abc")},
                    {"op": aerospike.OPERATOR_READ, "bin": "write_bin"},
                ],
                {"write_bin": ("a", "b", "c")},
            ),
            (
                ("test", "demo", 1),  # with_bin_bytearray
                [
                    {"op": aerospike.OPERATOR_PREPEND, "bin": bytearray("asd[;asjk", "utf-8"), "val": "ram"},
                    {"op": aerospike.OPERATOR_READ, "bin": bytearray("asd[;asjk", "utf-8")},
                ],
                {"asd[;asjk": "ram"},
            ),
            (
                ("test", "demo", "bytearray_key"),  # with_operator append_val bytearray
                [
                    {"op": aerospike.OPERATOR_APPEND, "bin": "bytearray_bin", "val": bytearray("abc", "utf-8")},
                    {"op": aerospike.OPERATOR_READ, "bin": "bytearray_bin"},
                ],
                {"bytearray_bin": bytearray("asd;as[d'as;dabc", "utf-8")},
            ),
            (
                ("test", "demo", "bytearray_new"),  # with_operator append_val bytearray_newrecord
                [
                    {
                        "op": aerospike.OPERATOR_APPEND,
                        "bin": "bytearray_bin",
                        "val": bytearray("asd;as[d'as;d", "utf-8"),
                    },
                    {"op": aerospike.OPERATOR_READ, "bin": "bytearray_bin"},
                ],
                {"bytearray_bin": bytearray("asd;as[d'as;d", "utf-8")},
            ),
            (
                ("test", "demo", "bytearray_key"),  # with_operatorprepend_valbytearray
                [
                    {"op": aerospike.OPERATOR_PREPEND, "bin": "bytearray_bin", "val": bytearray("abc", "utf-8")},
                    {"op": aerospike.OPERATOR_READ, "bin": "bytearray_bin"},
                ],
                {"bytearray_bin": bytearray("abcasd;as[d'as;d", "utf-8")},
            ),
            (
                ("test", "demo", "bytearray_new"),  # with_operatorprepend_valbytearray_newrecord
                [
                    {
                        "op": aerospike.OPERATOR_PREPEND,
                        "bin": "bytearray_bin",
                        "val": bytearray("asd;as[d'as;d", "utf-8"),
                    },
                    {"op": aerospike.OPERATOR_READ, "bin": "bytearray_bin"},
                ],
                {"bytearray_bin": bytearray("asd;as[d'as;d", "utf-8")},
            ),
        ],
    )
    def test_pos_operate_with_correct_paramters(self, key, llist, expected):
        """
        Invoke operate() with correct parameters
        """

        key, _, bins = self.as_connection.operate(key, llist)

        assert bins == expected
        self.as_connection.remove(key)

    def test_pos_operate_with_increment_positive_float_value(self):
        """
        Invoke operate() with correct parameters
        """
        if TestOperate.skip_old_server is True:
            pytest.skip("Server does not support increment on float type")
        key = ("test", "demo", 6)
        llist = [
            {"op": aerospike.OPERATOR_INCR, "bin": "age", "val": 3.5},
            {"op": aerospike.OPERATOR_READ, "bin": "age"},
        ]

        key, _, bins = self.as_connection.operate(key, llist)

        assert bins == {"age": 9.8}

    def test_pos_operate_with_correct_policy(self):
        """
        Invoke operate() with correct policy
        """
        key = ("test", "demo", 1)
        policy = {
            "timeout": 1000,
            "key": aerospike.POLICY_KEY_SEND,
            "commit_level": aerospike.POLICY_COMMIT_LEVEL_MASTER,
        }

        llist = [
            {"op": aerospike.OPERATOR_APPEND, "bin": "name", "val": "aa"},
            {"op": aerospike.OPERATOR_INCR, "bin": "age", "val": 3},
            {"op": aerospike.OPERATOR_READ, "bin": "name"},
        ]

        key, _, bins = self.as_connection.operate(key, llist, {}, policy)

        assert bins == {"name": "name1aa"}
        assert key == (
            "test",
            "demo",
            1,
            bytearray(b"\xb7\xf4\xb88\x89\xe2\xdag\xdeh>\x1d\xf6\x91\x9a\x1e\xac\xc4F\xc8"),
        )

        self.as_connection.remove(key)

    def test_pos_operate_with_policy_key_digest(self):
        """
        Invoke operate() with correct policy
        """
        key = ("test", "demo", None, bytearray("asd;as[d'as;djk;uyfl", "utf-8"))
        rec = {
            "name": "name%s" % (str(1)),
            "age": 1,
        }
        policy = {"timeout": 1000, "key": aerospike.POLICY_KEY_DIGEST}
        self.as_connection.put(key, rec)

        llist = [
            {"op": aerospike.OPERATOR_APPEND, "bin": "name", "val": "aa"},
            {"op": aerospike.OPERATOR_INCR, "bin": "age", "val": 3},
            {"op": aerospike.OPERATOR_READ, "bin": "name"},
        ]

        key, _, bins = self.as_connection.operate(key, llist, {}, policy)

        assert bins == {"name": "name1aa"}
        assert key == ("test", "demo", None, bytearray(b"asd;as[d'as;djk;uyfl"))
        self.as_connection.remove(key)

    @pytest.mark.parametrize(
        "key, policy, meta, llist",
        [
            (
                ("test", "demo", 1),
                {
                    "timeout": 1000,
                    "key": aerospike.POLICY_KEY_SEND,
                    "gen": aerospike.POLICY_GEN_IGNORE,
                    "commit_level": aerospike.POLICY_COMMIT_LEVEL_ALL,
                },
                {"gen": 10, "ttl": 1200},
                [
                    {"op": aerospike.OPERATOR_APPEND, "bin": "name", "val": "aa"},
                    {"op": aerospike.OPERATOR_INCR, "bin": "age", "val": 3},
                    {"op": aerospike.OPERATOR_READ, "bin": "name"},
                ],
            ),
        ],
    )
    def test_pos_operate_with_policy_gen_ignore(self, key, policy, meta, llist):
        """
        Invoke operate() with gen ignore.
        """

        key, meta, bins = self.as_connection.operate(key, llist, meta, policy)

        assert bins == {"name": "name1aa"}
        assert key == (
            "test",
            "demo",
            1,
            bytearray(b"\xb7\xf4\xb88\x89\xe2\xdag\xdeh>\x1d\xf6\x91\x9a\x1e\xac\xc4F\xc8"),
        )

    def test_pos_operate_with_policy_gen_EQ(self):
        """
        Invoke operate() with gen EQ positive.
        """
        key = ("test", "demo", 1)
        policy = {"timeout": 1000, "key": aerospike.POLICY_KEY_SEND, "gen": aerospike.POLICY_GEN_EQ}
        (key, meta) = self.as_connection.exists(key)
        gen = meta["gen"]
        meta = {"gen": gen, "ttl": 1200}

        llist = [
            {"op": aerospike.OPERATOR_APPEND, "bin": "name", "val": "aa"},
            {"op": aerospike.OPERATOR_INCR, "bin": "age", "val": 3},
            {"op": aerospike.OPERATOR_READ, "bin": "name"},
        ]

        (key, meta, bins) = self.as_connection.operate(key, llist, meta, policy)

        assert bins == {"name": "name1aa"}
        assert key == (
            "test",
            "demo",
            1,
            bytearray(b"\xb7\xf4\xb88\x89\xe2\xdag\xdeh>\x1d\xf6\x91\x9a\x1e\xac\xc4F\xc8"),
        )

    @pytest.mark.parametrize(
        "key, llist",
        [
            (("test", "demo", 1), [{"op": aerospike.OPERATOR_TOUCH, "val": 4000}]),
            (("test", "demo", 1), [{"op": aerospike.OPERATOR_TOUCH, "bin": "age", "val": 4000}]),
            (("test", "demo", 1), [{"op": aerospike.OPERATOR_TOUCH, "bin": "age"}]),
            (("test", "demo", 1), [{"op": aerospike.OPERATOR_TOUCH}]),
        ],
    )
    def test_pos_operate_touch_operation_with_bin_and_value_combination(self, key, llist):
        """
        Invoke operate() with touch value with bin and value combination.
        """

        self.as_connection.operate(key, llist)

        (key, meta) = self.as_connection.exists(key)

        assert meta["ttl"] is not None

    def test_pos_operate_with_policy_gen_EQ_not_equal(self):
        """
        Invoke operate() with gen not equal.
        """
        key = ("test", "demo", 1)
        policy = {"timeout": 1000, "key": aerospike.POLICY_KEY_SEND, "gen": aerospike.POLICY_GEN_EQ}

        (key, meta) = self.as_connection.exists(key)
        gen = meta["gen"]
        meta = {"gen": gen + 5, "ttl": 1200}
        llist = [
            {"op": aerospike.OPERATOR_APPEND, "bin": "name", "val": "aa"},
            {"op": aerospike.OPERATOR_INCR, "bin": "age", "val": 3},
            {"op": aerospike.OPERATOR_READ, "bin": "name"},
        ]
        try:
            key, meta, _ = self.as_connection.operate(key, llist, meta, policy)

        except e.RecordGenerationError as exception:
            assert exception.code == 3

        (key, meta, bins) = self.as_connection.get(key)
        assert bins == {"age": 1, "name": "name1"}
        assert key == (
            "test",
            "demo",
            None,
            bytearray(b"\xb7\xf4\xb88\x89\xe2\xdag\xdeh>\x1d\xf6\x91\x9a\x1e\xac\xc4F\xc8"),
        )

    def test_pos_operate_with_policy_gen_GT_lesser(self):
        """
        Invoke operate() with gen GT lesser.
        """
        key = ("test", "demo", 1)
        policy = {"timeout": 1000, "key": aerospike.POLICY_KEY_SEND, "gen": aerospike.POLICY_GEN_GT}
        (key, meta) = self.as_connection.exists(key)
        gen = meta["gen"]
        meta = {"gen": gen, "ttl": 1200}

        llist = [
            {"op": aerospike.OPERATOR_APPEND, "bin": "name", "val": "aa"},
            {"op": aerospike.OPERATOR_INCR, "bin": "age", "val": 3},
            {"op": aerospike.OPERATOR_READ, "bin": "name"},
        ]

        try:
            (key, meta, _) = self.as_connection.operate(key, llist, meta, policy)

        except e.RecordGenerationError as exception:
            assert exception.code == 3

        (key, meta, bins) = self.as_connection.get(key)
        assert bins == {"age": 1, "name": "name1"}
        assert key == (
            "test",
            "demo",
            None,
            bytearray(b"\xb7\xf4\xb88\x89\xe2\xdag\xdeh>\x1d\xf6\x91\x9a\x1e\xac\xc4F\xc8"),
        )

    def test_pos_operate_touch_with_meta(self):
        """
        Invoke operate() OPERATE_TOUCH using meta to pass in ttl.
        """
        key = ("test", "demo", 1)
        (key, _) = self.as_connection.exists(key)
        meta = {"ttl": 1200}

        llist = [{"op": aerospike.OPERATOR_TOUCH}]

        try:
            (key, meta, _) = self.as_connection.operate(key, llist, meta)

        except e.RecordGenerationError as exception:
            assert exception.code == 3

        (key, meta, bins) = self.as_connection.get(key)
        assert bins == {"age": 1, "name": "name1"}
        assert meta["ttl"] <= 1300
        assert meta["ttl"] > 800

    def test_pos_operate_with_policy_gen_GT(self):
        """
        Invoke operate() with gen GT positive.
        """
        key = ("test", "demo", 1)
        policy = {"timeout": 1000, "key": aerospike.POLICY_KEY_SEND, "gen": aerospike.POLICY_GEN_GT}
        (key, meta) = self.as_connection.exists(key)
        gen = meta["gen"]
        meta = {"gen": gen + 5, "ttl": 1200}

        llist = [
            {"op": aerospike.OPERATOR_APPEND, "bin": "name", "val": "aa"},
            {"op": aerospike.OPERATOR_INCR, "bin": "age", "val": 3},
            {"op": aerospike.OPERATOR_READ, "bin": "name"},
        ]

        (key, meta, bins) = self.as_connection.operate(key, llist, meta, policy)

        assert bins == {"name": "name1aa"}
        assert key == (
            "test",
            "demo",
            1,
            bytearray(b"\xb7\xf4\xb88\x89\xe2\xdag\xdeh>\x1d\xf6\x91\x9a\x1e\xac\xc4F\xc8"),
        )

    def test_pos_operate_with_nonexistent_key(self):
        """
        Invoke operate() with non-existent key
        """
        key1 = ("test", "demo", "key11")
        llist = [
            {"op": aerospike.OPERATOR_PREPEND, "bin": "loc", "val": "mumbai"},
            {"op": aerospike.OPERATOR_READ, "bin": "loc"},
        ]
        _, _, bins = self.as_connection.operate(key1, llist)

        assert bins == {"loc": "mumbai"}
        self.as_connection.remove(key1)

    def test_pos_operate_with_nonexistent_bin(self):
        """
        Invoke operate() with non-existent bin
        """
        key = ("test", "demo", 1)
        llist = [
            {"op": aerospike.OPERATOR_APPEND, "bin": "addr", "val": "pune"},
            {"op": aerospike.OPERATOR_READ, "bin": "addr"},
        ]
        key, _, bins = self.as_connection.operate(key, llist)

        assert bins == {"addr": "pune"}

    def test_pos_operate_increment_nonexistent_key(self):
        """
        Invoke operate() with increment with nonexistent_key
        """
        key = ("test", "demo", "non_existentkey")
        llist = [{"op": aerospike.OPERATOR_INCR, "bin": "age", "val": 5}]

        self.as_connection.operate(key, llist)

        (key, _, bins) = self.as_connection.get(key)

        assert bins == {"age": 5}

        self.as_connection.remove(key)

    def test_pos_operate_increment_nonexistent_bin(self):
        """
        Invoke operate() with increment with nonexistent_bin
        """
        key = ("test", "demo", 1)
        llist = [{"op": aerospike.OPERATOR_INCR, "bin": "my_age", "val": 5}]

        self.as_connection.operate(key, llist)

        (key, _, bins) = self.as_connection.get(key)

        assert bins == {"my_age": 5, "age": 1, "name": "name1"}

    def test_pos_operate_with_correct_paramters_without_connection(self):
        """
        Invoke operate() with correct parameters without connection
        """
        key = ("test", "demo", 1)
        config = TestBaseClass.get_connection_config()
        client1 = aerospike.client(config)
        client1.close()
        llist = [
            {"op": aerospike.OPERATOR_PREPEND, "bin": "name", "val": "ram"},
            {"op": aerospike.OPERATOR_INCR, "bin": "age", "val": 3},
            {"op": aerospike.OPERATOR_READ, "bin": "name"},
        ]

        try:
            key, _, _ = client1.operate(key, llist)

        except e.ClusterError as exception:
            assert exception.code == 11

    def test_pos_operate_write_set_to_aerospike_null(self):
        """
        Invoke operate() with write command with bin set to aerospike_null
        """
        key = ("test", "demo", "null_record")

        bins = {"name": "John", "no": 3}

        assert 0 == self.as_connection.put(key, bins)

        (key, _, bins) = self.as_connection.get(key)

        assert {"name": "John", "no": 3} == bins

        llist = [
            {"op": aerospike.OPERATOR_WRITE, "bin": "no", "val": aerospike.null()},
            {"op": aerospike.OPERATOR_READ, "bin": "no"},
        ]

        (key, _, bins) = self.as_connection.operate(key, llist)
        assert {} == bins

        self.as_connection.remove(key)

    @pytest.mark.parametrize(
        "key, llist, expected",
        [
            (
                ("test", "demo", "prepend_int"),  # prepend_with_int
                [
                    {"op": aerospike.OPERATOR_PREPEND, "bin": "age", "val": 4},
                    {"op": aerospike.OPERATOR_READ, "bin": "age"},
                ],
                {"age": 4},
            ),
            (
                ("test", "demo", "append_dict"),  # append_with_dict
                [
                    {"op": aerospike.OPERATOR_APPEND, "bin": "dict", "val": {"a": 1, "b": 2}},
                    {"op": aerospike.OPERATOR_READ, "bin": "dict"},
                ],
                {"dict": {"a": 1, "b": 2}},
            ),
            (
                ("test", "demo", "incr_string"),  # incr_with_string
                [
                    {"op": aerospike.OPERATOR_INCR, "bin": "name", "val": "aerospike"},
                    {"op": aerospike.OPERATOR_READ, "bin": "name"},
                ],
                {"name": "aerospike"},
            ),
        ],
    )
    def test_pos_operate_new_record(self, key, llist, expected):
        """
        Invoke operate() with prepend command on a new record
        """
        (key, _, bins) = TestOperate.client_no_typechecks.operate(key, llist)
        assert expected == bins
        TestOperate.client_no_typechecks.remove(key)

    @pytest.mark.parametrize(
        "key, llist",
        [
            (
                ("test", "demo", 1),
                [
                    {"op": aerospike.OPERATOR_PREPEND, "bin": "age", "val": 4},  # int
                    {"op": aerospike.OPERATOR_READ, "bin": "age"},
                ],
            ),
            (
                ("test", "demo", "existing_key"),  # Existing list
                [
                    {"op": aerospike.OPERATOR_PREPEND, "bin": "list", "val": ["c"]},
                    {"op": aerospike.OPERATOR_READ, "bin": "list"},
                ],
            ),
            (
                ("test", "demo", "existing_key"),  # Existing dict
                [
                    {"op": aerospike.OPERATOR_APPEND, "bin": "dict", "val": {"c": 2}},
                    {"op": aerospike.OPERATOR_READ, "bin": "dict"},
                ],
            ),
            (
                ("test", "demo", "existing_key"),  # Exiting float
                [
                    {"op": aerospike.OPERATOR_APPEND, "bin": "float", "val": 3.4},
                    {"op": aerospike.OPERATOR_READ, "bin": "float"},
                ],
            ),
            (
                ("test", "demo", 1),  # Existing string
                [
                    {"op": aerospike.OPERATOR_INCR, "bin": "name", "val": "aerospike"},
                    {"op": aerospike.OPERATOR_READ, "bin": "name"},
                ],
            ),
            (
                ("test", "demo", "existing_key"),  # Existing Bytearray
                [
                    {"op": aerospike.OPERATOR_INCR, "bin": "bytearray", "val": bytearray("abc", "utf-8")},
                    {"op": aerospike.OPERATOR_READ, "bin": "bytearray"},
                ],
            ),
        ],
    )
    def test_pos_operate_prepend_with_existing_record(self, key, llist):
        """
        Invoke operate() with prepend command on a existing record
        """
        try:
            (key, _, _) = TestOperate.client_no_typechecks.operate(key, llist)

        except e.BinIncompatibleType as exception:
            assert exception.code == 12

        TestOperate.client_no_typechecks.remove(key)

    def test_pos_operate_incr_with_geospatial_new_record(self):
        """
        Invoke operate() with incr command on a new record
        """
        key = ("test", "demo", "geospatial_key")

        llist = [
            {
                "op": aerospike.OPERATOR_INCR,
                "bin": "geospatial",
                "val": aerospike.GeoJSON({"type": "Point", "coordinates": [42.34, 58.62]}),
            },
            {"op": aerospike.OPERATOR_READ, "bin": "geospatial"},
        ]

        (key, _, bins) = TestOperate.client_no_typechecks.operate(key, llist)

        assert bins["geospatial"].unwrap() == {"coordinates": [42.34, 58.62], "type": "Point"}
        TestOperate.client_no_typechecks.remove(key)

    def test_pos_operate_with_bin_length_extra_nostricttypes(self):
        """
        Invoke operate() with bin length extra. Strict types disabled
        """
        key = ("test", "demo", 1)

        max_length = "a"
        for _ in range(20):
            max_length = max_length + "a"

        llist = [
            {"op": aerospike.OPERATOR_PREPEND, "bin": "name", "val": "ram"},
            {"op": aerospike.OPERATOR_INCR, "bin": max_length, "val": 3},
        ]

        TestOperate.client_no_typechecks.operate(key, llist)

        (key, _, bins) = TestOperate.client_no_typechecks.get(key)

        assert bins == {"name": "ramname1", "age": 1}

    def test_pos_operate_with_command_invalid_nostricttypes(self):
        """
        Invoke operate() with an invalid command. Strict types disabled
        """
        key = ("test", "demo", 1)

        llist = [
            {"op": aerospike.OPERATOR_PREPEND, "bin": "name", "val": "ram"},
            {"op": 3, "bin": "age", "val": 3},
            {"op": aerospike.OPERATOR_READ, "bin": "name"},
        ]

        key, _, bins = TestOperate.client_no_typechecks.operate(key, llist)

        assert bins == {"name": "ramname1"}

    def test_pos_operate_prepend_set_to_aerospike_null(self):
        """
        Invoke operate() with prepend command with bin set to aerospike_null
        """
        key = ("test", "demo", "null_record")

        bins = {"name": "John", "no": 3}

        assert 0 == self.as_connection.put(key, bins)

        (key, _, bins) = self.as_connection.get(key)

        assert {"name": "John", "no": 3} == bins

        llist = [
            {"op": aerospike.OPERATOR_PREPEND, "bin": "no", "val": aerospike.null()},
            {"op": aerospike.OPERATOR_READ, "bin": "no"},
        ]

        try:
            (key, _, bins) = self.as_connection.operate(key, llist)

        except e.InvalidRequest as exception:
            assert exception.code == 4
        self.as_connection.remove(key)

    @pytest.mark.parametrize(
        "list, result, bin, expected",
        [
            (
                [
                    {"op": aerospike.OP_LIST_APPEND, "bin": "int_bin", "val": 7},
                    {"op": aerospike.OP_LIST_GET, "bin": "int_bin", "index": 4},
                ],
                {"int_bin": 7},
                "int_bin",
                [1, 2, 3, 4, 7],
            ),
            (
                [
                    {"op": aerospike.OP_LIST_APPEND_ITEMS, "bin": "int_bin", "val": [7, 9]},
                    {"op": aerospike.OP_LIST_GET_RANGE, "bin": "int_bin", "index": 3, "val": 3},
                ],
                {"int_bin": [4, 7, 9]},
                "int_bin",
                [1, 2, 3, 4, 7, 9],
            ),
            (
                [
                    {"op": aerospike.OP_LIST_INSERT, "bin": "int_bin", "val": 7, "index": 2},
                    {"op": aerospike.OP_LIST_POP, "bin": "int_bin", "index": 2},
                ],
                {"int_bin": 7},
                "int_bin",
                [1, 2, 3, 4],
            ),
            (
                [
                    {"op": aerospike.OP_LIST_INSERT_ITEMS, "bin": "int_bin", "val": [7, 9], "index": 2},
                    {"op": aerospike.OP_LIST_POP_RANGE, "bin": "int_bin", "index": 2, "val": 2},
                ],
                {"int_bin": [7, 9]},
                "int_bin",
                [1, 2, 3, 4],
            ),
            (
                [
                    {"op": aerospike.OP_LIST_SET, "bin": "int_bin", "index": 2, "val": 18},
                    {"op": aerospike.OP_LIST_GET, "bin": "int_bin", "index": 2},
                ],
                {"int_bin": 18},
                "int_bin",
                [1, 2, 18, 4],
            ),
            (
                [
                    {"op": aerospike.OP_LIST_SET, "bin": "int_bin", "index": 6, "val": 10},
                    {"op": aerospike.OP_LIST_GET, "bin": "int_bin", "index": 6},
                ],
                {"int_bin": 10},
                "int_bin",
                [1, 2, 3, 4, None, None, 10],
            ),
        ],
    )
    def test_pos_operate_with_list_addition_operations(self, list, result, bin, expected):
        """
        Invoke operate() with list addition operations
        """
        key = ("test", "demo", "list_key")

        key, _, bins = self.as_connection.operate(key, list)

        assert bins == result

        key, _, bins = self.as_connection.get(key)

        assert bins[bin] == expected

    @pytest.mark.parametrize(
        "list, bin, expected",
        [
            (
                [
                    {"op": aerospike.OP_LIST_REMOVE, "bin": "int_bin", "index": 2},
                ],
                "int_bin",
                [1, 2, 4],
            ),
            (
                [
                    {"op": aerospike.OP_LIST_REMOVE_RANGE, "bin": "int_bin", "index": 2, "val": 2},
                ],
                "int_bin",
                [1, 2],
            ),
            (
                [
                    {"op": aerospike.OP_LIST_TRIM, "bin": "int_bin", "index": 2, "val": 2},
                ],
                "int_bin",
                [3, 4],
            ),
            ([{"op": aerospike.OP_LIST_CLEAR, "bin": "int_bin"}], "int_bin", []),
        ],
    )
    def test_pos_operate_with_list_remove_operations(self, list, bin, expected):
        """
        Invoke operate() with list remove operations
        """
        key = ("test", "demo", "list_key")

        self.as_connection.operate(key, list)

        key, _, bins = self.as_connection.get(key)

        assert bins[bin] == expected

    def test_pos_operate_with_list_size(self):
        """
        Invoke operate() with list_size operation
        """
        key = ("test", "demo", "list_key")
        list = [{"op": aerospike.OP_LIST_SIZE, "bin": "int_bin"}]

        key, _, bins = self.as_connection.operate(key, list)

        assert bins == {"int_bin": 4}

    def test_list_increment_with_valid_value(self):
        """
        previous list was [1, 2, 3, 4]
        new should be [1, 2, 23, 4]
        """
        key = ("test", "demo", "list_key")
        list = [{"op": aerospike.OP_LIST_INCREMENT, "bin": "int_bin", "index": 2, "val": 20}]

        _, _, bins = self.as_connection.operate(key, list)

        assert bins == {"int_bin": 23}
        _, _, bins = self.as_connection.get(key)

        assert bins["int_bin"] == [1, 2, 23, 4]

    def test_list_increment_with_missing_value(self):
        """
        previous list was [1, 2, 3, 4]
        new should be [1, 2, 23, 4]
        """
        key = ("test", "demo", "list_key")
        list = [
            {
                "op": aerospike.OP_LIST_INCREMENT,
                "bin": "int_bin",
                "index": 2,
            }
        ]

        with pytest.raises(e.ParamError):
            self.as_connection.operate(key, list)

    def test_list_increment_with_missing_index(self):
        """
        previous list was [1, 2, 3, 4]
        new should be [1, 2, 23, 4]
        """
        key = ("test", "demo", "list_key")
        list = [{"op": aerospike.OP_LIST_INCREMENT, "bin": "int_bin", "val": 20}]

        with pytest.raises(e.ParamError):
            self.as_connection.operate(key, list)

    def test_list_increment_with_missing_bin(self):
        """
        previous list was [1, 2, 3, 4]
        new should be [1, 2, 23, 4]
        """
        key = ("test", "demo", "list_key")
        list = [{"op": aerospike.OP_LIST_INCREMENT, "index": 2, "val": 20}]

        with pytest.raises(e.ParamError):
            self.as_connection.operate(key, list)

    def test_list_increment_with_incorrect_value_type(self):
        """
        previous list was [1, 2, 3, 4]
        new should be [1, 2, 23, 4]
        """
        key = ("test", "demo", "list_key")
        list = [{"op": aerospike.OP_LIST_INCREMENT, "index": 2, "bin": "int_bin", "val": "twenty"}]

        with pytest.raises(e.AerospikeError):
            self.as_connection.operate(key, list)

    def test_pos_operate_with_list_get_range_val_out_of_bounds(self):
        """
        Invoke operate() with list_get_range operation and value out of bounds
        """
        key = ("test", "demo", "list_key")
        list = [{"op": aerospike.OP_LIST_GET_RANGE, "bin": "int_bin", "index": 2, "val": 9}]

        (key, meta, bins) = self.as_connection.operate(key, list)

        assert bins == {"int_bin": [3, 4]}

    def test_pos_operate_with_list_trim_val_with_negative_value(self):
        """
        Invoke operate() with list_trimoperation and value is negative
        """
        key = ("test", "demo", "list_key")
        list = [{"op": aerospike.OP_LIST_TRIM, "bin": "int_bin", "index": 1, "val": -9}]

        (key, meta, bins) = self.as_connection.operate(key, list)

        (key, meta, bins) = self.as_connection.get(key)

        assert bins["int_bin"] == [2, 3, 4]

    def test_pos_operate_with_list_insert_index_negative(self):
        """
        Invoke operate() with list_insert and item index is a negative value
        """
        key = ("test", "demo", "list_key")
        list = [{"op": aerospike.OP_LIST_INSERT, "bin": "int_bin", "index": -2, "val": 9}]

        (key, meta, bins) = self.as_connection.operate(key, list)

        (key, meta, bins) = self.as_connection.get(key)

        assert bins["int_bin"] == [1, 2, 9, 3, 4]

    @pytest.mark.parametrize(
        "list, result, bin, expected",
        [
            (
                [
                    {"op": aerospike.OP_LIST_APPEND, "bin": "string_bin", "val": {"new_val": 1}},
                    {"op": aerospike.OP_LIST_GET, "bin": "string_bin", "index": 4},
                ],
                {"string_bin": {"new_val": 1}},
                "string_bin",
                ["a", "b", "c", "d", {"new_val": 1}],
            ),
            (
                [
                    {"op": aerospike.OP_LIST_APPEND_ITEMS, "bin": "string_bin", "val": [["z", "x"], ("y", "w")]},
                    {"op": aerospike.OP_LIST_GET_RANGE, "bin": "string_bin", "index": 3, "val": 3},
                ],
                {"string_bin": ["d", ["z", "x"], ("y", "w")]},
                "string_bin",
                ["a", "b", "c", "d", ["z", "x"], ("y", "w")],
            ),
            (
                [
                    {"op": aerospike.OP_LIST_INSERT, "bin": "string_bin", "val": True, "index": 2},
                    {"op": aerospike.OP_LIST_POP, "bin": "string_bin", "index": 2},
                ],
                {"string_bin": True},
                "string_bin",
                ["a", "b", "c", "d"],
            ),
            (
                [
                    {
                        "op": aerospike.OP_LIST_INSERT_ITEMS,
                        "bin": "string_bin",
                        "val": [bytearray("abc", "utf-8"), "xyz"],
                        "index": 2,
                    },
                    {"op": aerospike.OP_LIST_POP_RANGE, "bin": "string_bin", "index": 2, "val": 2},
                ],
                {"string_bin": [bytearray(b"abc"), "xyz"]},
                "string_bin",
                ["a", "b", "c", "d"],
            ),
        ],
    )
    def test_pos_operate_with_list_operations_different_datatypes(self, list, result, bin, expected):
        """
        Invoke operate() with list operations using different datatypes
        """
        key = ("test", "demo", "list_key")

        key, _, bins = self.as_connection.operate(key, list)

        assert bins == result

        key, _, bins = self.as_connection.get(key)

        assert bins[bin] == expected

    # Negative Tests
    def test_neg_operate_with_no_parameters(self):
        """
        Invoke opearte() without any mandatory parameters.
        """
        with pytest.raises(TypeError) as typeError:
            self.as_connection.operate()
        assert "argument 'key' (pos 1)" in str(typeError.value)

    def test_neg_operate_list_operation_bin_notlist(self):
        """
        Invoke operate() with a list operation and bin does not contain list
        """
        key = ("test", "demo", 1)
        list = [{"op": aerospike.OP_LIST_INSERT, "bin": "age", "index": 2, "val": 9}]

        try:
            (key, _, _) = self.as_connection.operate(key, list)

        except e.BinIncompatibleType as exception:
            assert exception.code == 12

    def test_neg_operate_append_items_not_a_list(self):
        """
        Invoke operate() with list addition operations negative
        """
        key = ("test", "demo", "list_key")

        list = [
            {"op": aerospike.OP_LIST_APPEND_ITEMS, "bin": "int_bin", "val": 7},
        ]

        try:
            key, _, bins = self.as_connection.operate(key, list)
        except e.ParamError as exception:
            assert exception.code == -2

    @pytest.mark.parametrize(
        "list",
        [
            ([{"op": aerospike.OP_LIST_GET, "bin": "int_bin", "index": 7}]),
            (
                [
                    {"op": aerospike.OP_LIST_CLEAR, "bin": "int_bin"},
                    {"op": aerospike.OP_LIST_POP, "bin": "int_bin", "index": 2},
                ]
            ),
            (
                [
                    {"op": aerospike.OP_LIST_CLEAR, "bin": "int_bin"},
                    {"op": aerospike.OP_LIST_REMOVE, "bin": "int_bin", "index": 2},
                ]
            ),
        ],
    )
    def test_neg_operate_list_invalid_requests(self, list):
        """
        Invoke operate() with list addition operations negative
        """
        key = ("test", "demo", "list_key")
        try:
            key, _, _ = self.as_connection.operate(key, list)
        except e.OpNotApplicable as exception:
            assert exception.code == 26

    def test_neg_operate_with_command_invalid(self):
        """
        Invoke operate() with an invalid command. Strict types enabled
        """
        key = ("test", "demo", 1)

        llist = [
            {"op": aerospike.OPERATOR_PREPEND, "bin": "name", "val": "ram"},
            {"op": 3, "bin": "age", "val": 3},
            {"op": aerospike.OPERATOR_READ, "bin": "name"},
        ]

        try:
            key, _, _ = self.as_connection.operate(key, llist)

        except e.ParamError as exception:
            assert exception.code == -2

    def test_neg_operate_with_bin_length_extra(self):
        """
        Invoke operate() with bin length extra. Strict types enabled
        """
        key = ("test", "demo", 1)

        max_length = "a"
        for _ in range(20):
            max_length = max_length + "a"

        llist = [
            {"op": aerospike.OPERATOR_PREPEND, "bin": "name", "val": "ram"},
            {"op": aerospike.OPERATOR_INCR, "bin": max_length, "val": 3},
            {"op": aerospike.OPERATOR_READ, "bin": "name"},
        ]

        try:
            key, _, _ = self.as_connection.operate(key, llist)

        except e.BinNameError as exception:
            assert exception.code == 21

    def test_neg_operate_empty_string_key(self):
        """
        Invoke operate() with empty string key
        """
        llist = [{"op": aerospike.OPERATOR_PREPEND, "bin": "name", "val": "ram"}]
        try:
            self.as_connection.operate("", llist)

        except e.ParamError as exception:
            assert exception.code == -2

    def test_neg_operate_with_extra_parameter(self):
        """
        Invoke operate() with extra parameter.
        """
        key = ("test", "demo", 1)
        policy = {"timeout": 1000}
        llist = [{"op": aerospike.OPERATOR_PREPEND, "bin": "name", "val": "ram"}]
        with pytest.raises(TypeError) as typeError:
            self.as_connection.operate(key, llist, {}, policy, "")

        assert "operate() takes at most 4 arguments (5 given)" in str(typeError.value)

    def test_neg_operate_policy_is_string(self):
        """
        Invoke operate() with policy is string
        """
        key = ("test", "demo", 1)
        llist = [{"op": aerospike.OPERATOR_PREPEND, "bin": "name", "val": "ram"}]
        try:
            self.as_connection.operate(key, llist, {}, "")

        except e.ParamError as exception:
            assert exception.code == -2

    def test_neg_operate_key_is_none(self):
        """
        Invoke operate() with key is none
        """
        llist = [{"op": aerospike.OPERATOR_PREPEND, "bin": "name", "val": "ram"}]
        try:
            self.as_connection.operate(None, llist)

        except e.ParamError as exception:
            assert exception.code == -2

    @pytest.mark.parametrize(
        "key, policy, llist, ex_code",
        [
            (
                ("test", "demo", 1),
                {"timeout": 1000},
                [
                    {"op": aerospike.OPERATOR_APPEND, "bin": "name"},
                    {"op": aerospike.OPERATOR_INCR, "bin": "age", "val": 3},
                ],
                -2,
            ),
            (
                ("test", "demo", 1),
                {"timeout": 1000},
                [
                    {"op": aerospike.OPERATOR_APPEND, "bin": "name", "val": 3, "aa": 89},
                ],
                -2,
            ),
            (
                ("test", "demo", 1),  # with_incr_value_string
                {
                    "timeout": 1000,
                    "key": aerospike.POLICY_KEY_SEND,
                    "commit_level": aerospike.POLICY_COMMIT_LEVEL_MASTER,
                },
                [
                    {"op": aerospike.OPERATOR_INCR, "bin": "age", "val": "3"},
                    {"op": aerospike.OPERATOR_READ, "bin": "age"},
                ],
                -2,
            ),
        ],
    )
    def test_neg_operate_append_without_value_parameter(self, key, policy, llist, ex_code):
        """
        Invoke operate() with append operation and append val is not given
        """

        try:
            self.as_connection.operate(key, llist, {}, policy)

        except e.ParamError as exception:
            assert exception.code == ex_code

    def test_neg_operate_append_value_integer(self):
        """
        Invoke operate() with append value is of type integer
        """
        key = ("test", "demo", 1)
        llist = [
            {"op": aerospike.OPERATOR_APPEND, "bin": "name", "val": 12},
            {"op": aerospike.OPERATOR_INCR, "bin": "age", "val": 3},
            {"op": aerospike.OPERATOR_READ, "bin": "name"},
        ]

        try:
            self.as_connection.operate(key, llist)
        except e.ParamError as exception:
            assert exception.code == -2

    def test_neg_opearte_with_incorrect_polic(self):
        """
        Invoke operate() with incorrect policy
        """
        key = ("test", "demo", 1)
        policy = {"timeout": 0.5}
        llist = [
            {"op": aerospike.OPERATOR_PREPEND, "bin": "name", "val": "ram"},
            {"op": aerospike.OPERATOR_INCR, "bin": "age", "val": 3},
            {"op": aerospike.OPERATOR_READ, "bin": "name"},
        ]

        try:
            self.as_connection.operate(key, llist, {}, policy)

        except e.ParamError as exception:
            assert exception.code == -2

    def test_neg_opearte_on_same_bin(self):
        """
        Invoke operate() on same bin
        """
        key = ("test", "demo", 1)
        policy = {"timeout": 5000}
        llist = [
            {"op": aerospike.OPERATOR_PREPEND, "bin": "name", "val": "ram"},
            {"op": aerospike.OPERATOR_APPEND, "bin": "name", "val": "aa"},
            {"op": aerospike.OPERATOR_INCR, "bin": "age", "val": 3},
            {"op": aerospike.OPERATOR_READ, "bin": "name"},
        ]

        try:
            self.as_connection.operate(key, llist, {}, policy)

        except e.InvalidRequest as exception:
            assert exception.code == 4



# -*- coding: utf-8 -*-
##########################################################################
# Copyright 2013-2021 Aerospike, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##########################################################################

import asyncio
import sys
import aerospike
import time
from aerospike_helpers.awaitable import io

from optparse import OptionParser

##########################################################################
# Options Parsing
##########################################################################

usage = "usage: %prog [options]"

optparser = OptionParser(usage=usage, add_help_option=False)

optparser.add_option(
    "--help", dest="help", action="store_true",
    help="Displays this message.")

optparser.add_option(
    "-U", "--username", dest="username", type="string",  default="ram", metavar="<USERNAME>",
    help="Username to connect to database.")

optparser.add_option(
    "-P", "--password", dest="password", type="string", default="ram", metavar="<PASSWORD>",
    help="Password to connect to database.")

optparser.add_option(
    "-h", "--host", dest="host", type="string", default="as-s1.as-network.com", metavar="<ADDRESS>",
    help="Address of Aerospike server.")

optparser.add_option(
    "-p", "--port", dest="port", type="int", default=3000, metavar="<PORT>",
    help="Port of the Aerospike server.")

optparser.add_option(
    "--timeout", dest="timeout", type="int", default=1000, metavar="<MS>",
    help="Client timeout")

optparser.add_option(
    "-n", "--namespace", dest="namespace", type="string", default="test", metavar="<NS>",
    help="Namespace of database.")

optparser.add_option(
    "-s", "--set", dest="set", type="string", default="demo", metavar="<SET>",
    help="Set to use within namespace of database.")

optparser.add_option(
    "-c", "--test_count", dest="test_count", type="int", default=128, metavar="<TEST_COUNT>",
    help="Number of async IO to spawn.")

(options, args) = optparser.parse_args()

if options.help:
    optparser.print_help()
    print()
    sys.exit(1)

##########################################################################
# Client Configuration
##########################################################################

config = {
    'hosts': [(options.host, options.port)],
    'policies': {
        'total_timeout': options.timeout
    }
}

##########################################################################
# Application
##########################################################################

exitCode = 0

try:

    # ----------------------------------------------------------------------------
    # Connect to Cluster
    # ----------------------------------------------------------------------------
    print(f"Connecting to {options.host}:{options.port} with {options.username}:{options.password}")
    client = aerospike.client(config).connect(
        options.username, options.password)

    # ----------------------------------------------------------------------------
    # Perform Operation
    # ----------------------------------------------------------------------------

    try:
        io_results = {}
        test_count = options.test_count
        namespace = options.namespace if options.namespace and options.namespace != 'None' else None
        set = options.set if options.set and options.set != 'None' else None
        policy = {
            'total_timeout': options.timeout
        }
        meta = None

        print(f"IO async test count:{test_count}")
        
        async def async_io(namespace, set, i):
            key = (namespace, \
                    set, \
                    str(i), \
                    client.get_key_digest(namespace, set, str(i)))
            context = {'result': {}}
            io_results[key[2]] = context
            result = None
            try:
                result = await io.get(client, key, policy)
            except Exception as eargs:
                print(f"error: {eargs.code}, {eargs.msg}, {eargs.file}, {eargs.line}")
            print(result)
            io_results[key[2]]['result'] = result
        async def main():
            func_list = []
            for i in range(test_count):
                func_list.append(async_io(namespace, set, i))
            await asyncio.gather(*func_list)
        asyncio.get_event_loop().run_until_complete(main())
        print(io_results)
        print(f"get_async completed with returning {len(io_results)} records")
    except Exception as e:
        print(f"error: {0} ".format(e), file=sys.stderr)
        rc = 1

    # ----------------------------------------------------------------------------
    # Close Connection to Cluster
    # ----------------------------------------------------------------------------

    client.close()

except Exception as eargs:
    print("error: {0}".format(eargs), file=sys.stderr)
    exitCode = 3

##########################################################################
# Exit
##########################################################################

sys.exit(exitCode)



# -*- coding: utf-8 -*-
import pytest
from aerospike import exception as e
from aerospike import predicates as p
from .test_base_class import TestBaseClass

import aerospike


def add_stream_udf(client):
    client.udf_put("stream_example.lua", 0)


def remove_stream_udf(client):
    client.udf_remove("stream_example.lua")


def add_required_index(client):
    client.index_integer_create("test", "demo", "test_age", "age_index")


def remove_index(client):
    client.index_remove("test", "age_index")


@pytest.mark.xfail(reason="file permissions can cause this to fail")
class TestAggregate(object):
    def setup_class(cls):
        cls.connection_setup_functions = (add_required_index, add_stream_udf)

        cls.connection_teardown_functions = (remove_index, remove_stream_udf)

    @pytest.fixture(autouse=True)
    def setup(self, request, connection_with_config_funcs):
        """
        Setup Method
        """

        as_connection = connection_with_config_funcs
        for i in range(5):
            key = ("test", "demo", i)
            rec = {"name": "name%s" % (str(i)), "addr": "name%s" % (str(i)), "test_age": i, "no": i}
            as_connection.put(key, rec)

        def teardown():
            """
            Teardown Method
            """
            for i in range(5):
                key = ("test", "demo", i)
                as_connection.remove(key)

        request.addfinalizer(teardown)

    def test_pos_aggregate_with_where_bool_value(self):
        """
        Invoke aggregate() with where is bool value
        """
        query = self.as_connection.query("test", "demo")
        query.select("name", "test_age")
        query.where(p.between("test_age", True, True))
        query.apply("stream_example", "count")
        records = []

        def user_callback(value):
            records.append(value)

        query.foreach(user_callback)
        assert records[0] == 1

    def test_pos_aggregate_with_where_equals_value(self):
        """
        Invoke aggregate() with where is equal
        """
        query = self.as_connection.query("test", "demo")
        query.select("name", "test_age")
        query.where(p.equals("test_age", 2))
        query.apply("stream_example", "count")
        records = []

        def user_callback(value):
            records.append(value)

        query.foreach(user_callback)
        assert records[0] == 1

    def test_pos_aggregate_with_empty_module_function(self):
        """
        Invoke aggregate() with empty module and function
        """
        query = self.as_connection.query("test", "demo")
        query.select("name", "test_age")
        query.where(p.between("test_age", 1, 5))
        query.apply("", "")

        result = None

        def user_callback(value):
            _ = value

        query.foreach(user_callback)
        assert result is None

    def test_pos_aggregate_with_correct_parameters(self):
        """
        Invoke aggregate() with correct arguments
        """
        query = self.as_connection.query("test", "demo")
        query.select("name", "test_age")
        query.where(p.between("test_age", 1, 5))
        query.apply("stream_example", "count")

        records = []

        def user_callback(value):
            records.append(value)

        query.foreach(user_callback)
        assert records[0] == 4

    def test_pos_aggregate_with_policy(self):
        """
        Invoke aggregate() with policy
        """
        policy = {"timeout": 1000}
        query = self.as_connection.query("test", "demo")
        query.select("name", "test_age")
        query.where(p.between("test_age", 1, 5))
        query.apply("stream_example", "count")

        records = []

        def user_callback(value):
            records.append(value)

        query.foreach(user_callback, policy)
        assert records[0] == 4

    def test_pos_aggregate_with_extra_parameters_to_lua(self):
        """
        Invoke aggregate() with extra arguments
        """
        query = self.as_connection.query("test", "demo")
        query.select("name", "test_age")
        query.where(p.between("test_age", 1, 5))
        stream = None
        query.apply("stream_example", "count", [stream])

        records = []

        def user_callback(value):
            records.append(value)

        query.foreach(user_callback)
        assert records[0] == 4

    def test_pos_aggregate_with_extra_parameter_in_lua(self):
        """
        Invoke aggregate() with extra parameter in lua
        """
        query = self.as_connection.query("test", "demo")
        query.select("name", "test_age")
        query.where(p.between("test_age", 1, 5))
        query.apply("stream_example", "count_extra")

        records = []

        def user_callback(value):
            records.append(value)

        query.foreach(user_callback)
        assert records[0] == 4

    def test_pos_aggregate_with_arguments_to_lua_function(self):
        """
        Invoke aggregate() with unicode arguments to lua function.
        """
        query = self.as_connection.query("test", "demo")
        query.where(p.between("test_age", 0, 5))
        query.apply("stream_example", "group_count", ["name", "addr"])

        rec = []

        def callback(value):
            rec.append(value)

        query.foreach(callback)
        assert rec == [{"name4": 1, "name2": 1, "name3": 1, "name0": 1, "name1": 1}]

    def test_pos_aggregate_with_unicode_module_and_function_name(self):
        """
        Invoke aggregate() with unicode module and function names
        """
        query = self.as_connection.query("test", "demo")
        query.select("name", "test_age")
        query.where(p.between("test_age", 1, 5))
        query.apply("stream_example", "count")

        records = []

        def user_callback(value):
            records.append(value)

        query.foreach(user_callback)
        assert records[0] == 4

    def test_pos_aggregate_with_multiple_foreach_on_same_query_object(self):
        """
        Invoke aggregate() with multiple foreach on same query object.
        """
        query = self.as_connection.query("test", "demo")
        query.select("name", "test_age")
        query.where(p.between("test_age", 1, 5))
        query.apply("stream_example", "count")

        records = []

        def user_callback(value):
            records.append(value)

        query.foreach(user_callback)
        assert records[0] == 4

        records = []
        query.foreach(user_callback)
        assert records[0] == 4

    def test_pos_aggregate_with_multiple_results_call_on_same_query_object(self):
        """
        Invoke aggregate() with multiple foreach on same query object.
        """
        query = self.as_connection.query("test", "demo")
        query.select("name", "test_age")
        query.where(p.between("test_age", 1, 5))
        query.apply("stream_example", "count")

        records = query.results()
        assert records[0] == 4

        records = []
        records = query.results()
        assert records[0] == 4

    @pytest.mark.skip(reason="This function does not exist")
    def test_neg_aggregate_with_arguments_to_lua_function_having_float_value(self):
        """
        Invoke aggregate() with unicode arguments to lua function having a
        float value
        """
        query = self.as_connection.query("test", "demo")
        query.where(p.between("test_age", 0, 5))
        query.apply("stream_example", "double_group_count", ["name", "addr", 2.5])

        rec = []

        def callback(value):
            rec.append(value)

        try:
            query.foreach(callback)

        except e.ClientError as exception:
            assert exception.code == -1

    def test_neg_aggregate_with_correct_parameters_without_connection(self):
        """
        Invoke aggregate() with correct arguments without connection
        """
        config = TestBaseClass.get_connection_config()
        client1 = aerospike.client(config)
        client1.close()

        try:
            query = client1.query("test", "demo")
            query.select("name", "test_age")
            query.where(p.between("test_age", 1, 5))
            query.apply("stream_example", "count")

            records = []

            def user_callback(value):
                records.append(value)

            query.foreach(user_callback)

        except e.ClusterError as exception:
            assert exception.code == 11

    def test_neg_aggregate_with_extra_parameter(self):
        """
        Invoke aggregate() with extra parameter
        """
        policy = {"timeout": 1000}

        with pytest.raises(TypeError) as typeError:
            query = self.as_connection.query("test", "demo")
            query.select("name", "test_age")
            query.where(p.between("test_age", 1, 5))
            query.apply("stream_example", "count")

            def user_callback(value):
                _ = value

            query.foreach(user_callback, policy, "")

        assert "foreach() takes at most 2 arguments (3 given)" in str(typeError.value)

    def test_neg_aggregate_with_no_parameters(self):
        """
        Invoke aggregate() without any mandatory parameters.
        """
        try:
            query = self.as_connection.query()
            query.select()
            query.where()

        except e.ParamError as exception:
            assert exception.code == -2

    def test_neg_aggregate_no_sec_index(self):
        """
        Invoke aggregate() with no secondary index
        """
        try:
            query = self.as_connection.query("test", "demo")
            query.select("name", "no")
            query.where(p.between("no", 1, 5))
            query.apply("stream_example", "count")

            def user_callback(value):
                _ = value

            query.foreach(user_callback)
        except e.IndexNotFound as exception:
            assert exception.code == 201

    def test_neg_aggregate_with_incorrect_ns_set(self):
        """
        Invoke aggregate() with incorrect ns and set
        """
        try:
            query = self.as_connection.query("test1", "demo1")
            query.select("name", "test_age")
            query.where(p.equals("test_age", 1))
            query.apply("stream_example", "count")

            def user_callback(value):
                _ = value

            query.foreach(user_callback)

        except e.InvalidRequest as exception:
            assert exception.code == 4
        except e.NamespaceNotFound as exception:
            assert exception.code == 20

    def test_neg_aggregate_with_where_incorrect(self):
        """
        Invoke aggregate() with where is incorrect
        """
        query = self.as_connection.query("test", "demo")
        query.select("name", "test_age")
        query.where(p.equals("test_age", 165))
        query.apply("stream_example", "count")
        records = []

        def user_callback(value):
            records.append(value)

        query.foreach(user_callback)
        assert records == []

    def test_neg_aggregate_with_where_none_value(self):
        """
        Invoke aggregate() with where is null value
        """
        query = self.as_connection.query("test", "demo")
        query.select("name", "test_age")
        try:
            query.where(p.equals("test_age", None))
            query.apply("stream_example", "count")

            def user_callback(value):
                _ = value

            query.foreach(user_callback)

        except e.ParamError as exception:
            assert exception.code == -2



import pytest
from aerospike import exception as e


def get_key_with_digest_only(key):
    """
    Takes a key tuple, and removes a key tuple
    without a primary key, but with a key digest
    """
    return (key[0], key[1], None, key[3])


@pytest.mark.usefixtures("as_connection")
class TestKV(object):
    """
    Tests for fetching items by their keys
    """

    def setup_method(self):
        self.record_count = 0

    def test_using_a_single_key(self):
        """
        Using a single key,
        """
        key = ("test", "unittest", "1")

        # cleanup records

        self.record_count = 0

        def count_records(tuple):
            self.record_count += 1

        def remove_record(input_tuple):
            key, _, _ = input_tuple
            self.as_connection.remove(key)

        self.as_connection.scan("test", "unittest").foreach(remove_record)

        record_to_insert = {
            "i": 1234,
            "s": "abcd",
            "b": bytearray("efgh", "utf-8"),
            "l": [1357, "aceg", bytearray("aceg", "utf-8"), [1, 3, 5, 7], {"a": 1, "c": 3, "e": 5, "g": 7}],
            "m": {
                "i": 2468,
                "s": "bdfh",
                "l": [2468, "bdfh", bytearray("bdfh", "utf-8")],
                "m": {"b": 2, "d": 4, "f": 6, "h": 8},
            },
        }
        # create the record and insert it
        status = self.as_connection.put(key, record_to_insert)
        assert status == 0, "wrong return code"

        # ensure the record exists
        (key, meta) = self.as_connection.exists(key)
        assert meta is not None

        # count inserted records
        self.as_connection.scan("test", "unittest").foreach(count_records)
        assert self.record_count == 1

        # extract a record
        (key, meta, extracted_record) = self.as_connection.get(key)
        assert record_to_insert == extracted_record, "records do not match"

        # add a new bin with value to the record
        status = self.as_connection.put(key, {"hello": "world"})
        assert status == 0, "wrong return code"

        # add the same values to our local version of the record
        record_to_insert["hello"] = "world"

        # get the stored record
        (key, meta, extracted_record) = self.as_connection.get(key)
        assert record_to_insert == extracted_record, "records do not match"

        # remove the record from the DB
        status = self.as_connection.remove(key)
        assert status == 0, "wrong return code"

        # ensure not existent
        try:
            (key, meta) = self.as_connection.exists(key)
            """
            We are making the api backward compatible. In case of
            RecordNotFound an exception will not be raised.
            Instead Ok response is returned withe the
            meta as None. This might change with further releases.
            """
            assert meta is None
        except e.RecordNotFound as exception:
            assert exception.code == 2

        # count records
        self.record_count = 0
        self.as_connection.scan("test", "unittest").foreach(count_records)
        assert self.record_count == 0, "set should be empty"

    def test_with_single_key(self):
        """
        Using a single key, with digest only.
        """

        self.record_count = 0

        def count_records(tuple):
            self.record_count += 1

        key = ("test", "unittest", "1")

        # cleanup records
        def each_record(input_tuple):
            key, _, _ = input_tuple
            self.as_connection.remove(key)

        self.as_connection.scan("test", "unittest").foreach(each_record)

        record_to_insert = {
            "i": 1234,
            "s": "abcd",
            "b": bytearray("efgh", "utf-8"),
            "l": [1357, "aceg", bytearray("aceg", "utf-8"), [1, 3, 5, 7], {"a": 1, "c": 3, "e": 5, "g": 7}],
            "m": {
                "i": 2468,
                "s": "bdfh",
                "l": [2468, "bdfh", bytearray("bdfh", "utf-8")],
                "m": {"b": 2, "d": 4, "f": 6, "h": 8},
            },
            "a": {"aa": "11"},
            "k": {"kk": "22"},
        }

        # create the record
        status = self.as_connection.put(key, record_to_insert)
        assert status == 0, "wrong return code"

        # ensure existence
        (key, meta) = self.as_connection.exists(key)
        assert meta is not None

        # count records
        self.record_count = 0
        self.as_connection.scan("test", "unittest").foreach(count_records)
        assert self.record_count == 1, "set should have 1 record"

        # read it
        (key, meta, extracted_record) = self.as_connection.get(get_key_with_digest_only(key))
        assert record_to_insert == extracted_record, "records do not match"

        # Add a bin to the stored record
        status = self.as_connection.put(get_key_with_digest_only(key), {"hello": "world"})
        assert status == 0, "wrong return code"

        # Add a new value to the local record which matches the one
        # we stored remotely
        record_to_insert["hello"] = "world"

        # fetch the remote key and verify it matches the local version
        (key, meta, extracted_record) = self.as_connection.get(get_key_with_digest_only(key))
        assert record_to_insert == extracted_record, "records do not match"

        # remove it
        status = self.as_connection.remove(get_key_with_digest_only(key))
        assert status == 0, "wrong return code"

        # ensure not existent
        try:
            (key, meta) = self.as_connection.exists(get_key_with_digest_only(key))
            # We are making the api backward compatible. In case of
            # RecordNotFound an exception will not be raised.
            # Instead Ok response is returned withe the
            # meta as None. This might change with further releases.
            assert meta is None
        except e.RecordNotFound as exception:
            assert exception.code == 2

        # count records
        self.record_count = 0
        self.as_connection.scan("test", "unittest").foreach(count_records)
        assert self.record_count == 0

    @pytest.mark.skip(reason="Fails intermittently due to nature of callbacks")
    def test_with_multiple_keys(self):
        """
        Using multiple keys
        """
        from aerospike import predicates as p

        records = []

        def count_records_false(tuple):
            records.append(1)
            return False

        for i in range(3):
            key = ("test", "unittest", i)
            rec = {"name": "name%s" % (str(i)), "addr": "name%s" % (str(i)), "age": i, "no": i}
            self.as_connection.put(key, rec)

        self.as_connection.index_integer_create("test", "unittest", "age", "age_index", {})

        query = self.as_connection.query("test", "unittest")

        query.select("name", "age")
        self.record_count = 0
        query.where(p.between("age", 1, 3))

        query.foreach(count_records_false)
        self.record_count = len(records)
        # There are two records with age > 1 and < 3, but our callback
        # returns false, so iteration is stopped after the first
        # call
        assert self.record_count == 1, "foreach failed"

        for i in range(3):
            key = ("test", "unittest", i)
            self.as_connection.remove(key)

        self.as_connection.index_remove("test", "age_index", {})



from aerospike_helpers.operations import operations as op

# Insert 3 records
keys = [("test", "demo", f"employee{i}") for i in range(1, 4)]
bins = [
    {"id": 100, "balance": 200},
    {"id": 101, "balance": 400},
    {"id": 102, "balance": 300}
]
for key, bin in zip(keys, bins):
    client.put(key, bin)

# Increment ID by 100 and balance by 500 for all employees
# NOTE: batch_operate ops must include a write op
# get_batch_ops or get_many can be used for all read ops use cases.
ops = [
    op.increment("id", 100),
    op.increment("balance", 500),
    op.read("balance")
]

batchRecords = client.batch_operate(keys, ops)
print(batchRecords.result)
# 0

# Print each individual transaction's results
# and record if it was read from
for batchRecord in batchRecords.batch_records:
    print(f"{batchRecord.result}: {batchRecord.record}")
# 0: (('test', 'demo', 'employee1', bytearray(b'...')), {'ttl': 2592000, 'gen': 2}, {'id': None, 'balance': 700})
# 0: (('test', 'demo', 'employee2', bytearray(b'...')), {'ttl': 2592000, 'gen': 2}, {'id': None, 'balance': 900})
# 0: (('test', 'demo', 'employee3', bytearray(b'...')), {'ttl': 2592000, 'gen': 2}, {'id': None, 'balance': 800})



# -*- coding: utf-8 -*-
from aerospike_helpers.batch import records as br
from .test_base_class import TestBaseClass


class TestBatchRecords(TestBaseClass):
    def test_batch_read_all_bins_pos(self):
        """
        Test that batch Reads will allow read_all_bins with no ops
        """

        b = br.Read(("test", "demo", 1), ops=None, read_all_bins=True)

        assert b.read_all_bins

    def test_batch_default_arg_pos(self):
        """
        Test that batch Reads will always use a new batch_records list.
        """

        b = br.Read(("test", "demo", 1), ops=None)

        bwr = br.BatchRecords()

        bwr.batch_records.append(b)

        assert len(bwr.batch_records) == 1

        bwr = br.BatchRecords()

        assert len(bwr.batch_records) == 0



# -*- coding: utf-8 -*-

import pytest
import asyncio

try:
    pass
except Exception:
    pass
from . import test_data

# from collections import OrderedDict

from .test_base_class import TestBaseClass
import aerospike
from aerospike import exception as e
from aerospike_helpers.awaitable import io


class SomeClass(object):
    pass


aerospike.init_async()


@pytest.mark.usefixtures("as_connection")
class TestGet:
    @pytest.mark.parametrize("_input, _expected", test_data.pos_data)
    @pytest.mark.asyncio
    async def test_pos_get_put_with_key(self, _input, _expected, put_data):
        """
        Invoke get() with a key and not policy's dict.
        """
        put_data(self.as_connection, _input, _expected)

        @pytest.mark.asyncio
        async def async_io(_input):
            _, _, bin = await io.get(self.as_connection, _input)
            assert bin == _expected

        await asyncio.gather(async_io(_input))

    @pytest.mark.asyncio
    async def test_pos_get_with_data_missing(self):
        """
        Invoke get with a record which does not exist.
        """
        key = ("test", "demo", "non-existent-key-that-does-not-exist")

        @pytest.mark.asyncio
        async def async_io(io_input):
            with pytest.raises(e.RecordNotFound):
                await io.get(self.as_connection, io_input)

        await asyncio.gather(async_io(key))

    @pytest.mark.skip(reason="byte key not currently handled")
    @pytest.mark.asyncio
    async def test_get_information_using_bytes_key(self):
        record = {"bytes": "are_cool"}
        key = ("test", "demo", b"\x83X\xb5\xde")
        self.as_connection.put(key, record)

        @pytest.mark.asyncio
        async def async_io(io_input):
            _, _, rec_bin = await io.get(self.as_connection, io_input)
            assert rec_bin == record

        await asyncio.gather(async_io(key))
        self.as_connection.remove(key)

    @pytest.mark.asyncio
    async def test_pos_get_initkey_with_digest(self, put_data):
        """
        Invoke get() for a record having bytestring data.
        """
        key = ("test", "demo", None, bytearray("asd;as[d'as;djk;uyfl", "utf-8"))
        rec = {"name": "john"}

        policy = {"key": aerospike.POLICY_KEY_DIGEST}

        put_data(self.as_connection, key, rec, policy)

        async def async_io(key_input, policy_input=None):
            key, _, bins = await io.get(self.as_connection, key_input, policy_input)
            assert bins == {"name": "john"}
            assert key == ("test", "demo", None, bytearray(b"asd;as[d'as;djk;uyfl"))

        await asyncio.gather(async_io(key, policy))

    @pytest.mark.asyncio
    @pytest.mark.parametrize(
        "_input, _record, _policy, _expected",
        [
            (("test", "demo", 3), {"name": "name%s" % (str(3)), "age": 3}, aerospike.POLICY_KEY_DIGEST, None),
            (("test", "demo", 3), {"name": "name%s" % (str(3)), "age": 3}, aerospike.POLICY_KEY_SEND, 3),
        ],
    )
    async def test_pos_get_with_policy_key_digest(self, _input, _record, _policy, _expected, put_data):
        """
        Invoke get() for a record with POLICY_KEY_DIGEST
        """
        put_data(self.as_connection, _input, _record)

        async def async_io(key_input, policy_input=None):
            key, _, _ = await io.get(self.as_connection, key_input, policy_input)
            assert key[2] == _expected

        policy = {"key": _policy}
        await asyncio.gather(async_io(_input, policy))

    @pytest.mark.asyncio
    async def test_pos_get_initkey_with_policy_send(self, put_data):
        """
        Invoke get() for a record having string data.
        """

        key = ("test", "demo", 1)

        rec = {"name": "john", "age": 1}

        policy = {"key": aerospike.POLICY_KEY_SEND}

        put_data(self.as_connection, key, rec, policy)

        async def async_io(key_input, policy_input=None):
            key, _, bins = await io.get(self.as_connection, key_input, policy_input)
            assert bins == {"name": "john", "age": 1}
            assert key == (
                "test",
                "demo",
                1,
                bytearray(b"\xb7\xf4\xb88\x89\xe2\xdag\xdeh>\x1d\xf6\x91\x9a\x1e\xac\xc4F\xc8"),
            )

        await asyncio.gather(async_io(key, policy))

    # Negative get tests
    @pytest.mark.asyncio
    async def test_neg_get_with_no_parameter(self):
        """
        Invoke get() without any mandatory parameters.
        """

        async def async_io(key_input=None, policy_input=None):
            with pytest.raises(e.ParamError):
                await io.get(self.as_connection)

        await asyncio.gather(async_io())

    def test_neg_get_with_extra_parameter_in_key(self, put_data):
        """
        Invoke get() with extra parameter
        """
        key = ("test", "demo", None, bytearray("asd;as[d'as;djk;uyfl", "utf-8"), 1)
        rec = {"name": "john"}

        policy = {"key": aerospike.POLICY_KEY_DIGEST}
        with pytest.raises(e.ParamError):
            put_data(self.as_connection, key, rec, policy)

    @pytest.mark.asyncio
    async def test_neg_get_with_key_digest(self):
        """
        Invoke get() with a key digest.
        """
        key = ("test", "demo", 1)

        async def async_io(key_input=None, policy_input=None):
            try:
                key, _ = self.as_connection.exists(key_input)
                key, _, _ = await io.get(self.as_connection, (key[0], key[1], None, key[2]))
            except e.ParamError as exception:
                assert exception.code == -2
                assert exception.msg == "digest is invalid. expected a bytearray"
            except e.RecordNotFound as exception:
                assert exception.code == 2

        await asyncio.gather(async_io((key[0], key[1], None, key[2])))

    @pytest.mark.parametrize("key, ex_code, ex_msg", test_data.key_neg)
    @pytest.mark.asyncio
    async def test_neg_get_with_none(self, key, ex_code, ex_msg):
        """
        Invoke get() with None namespace/key in key tuple.
        """

        async def async_io(key_input=None, policy_input=None):
            with pytest.raises(e.ParamError):
                await io.get(self.as_connection, key_input)

        await asyncio.gather(async_io(key))

    @pytest.mark.asyncio
    async def test_neg_get_with_invalid_record(self):
        """
        Get a record which does not exist
        """
        key = ("test", "demo", "A PRIMARY KEY WHICH DOES NOT EXIST")

        async def async_io(key_input=None, policy_input=None):
            with pytest.raises(e.RecordNotFound):
                await io.get(self.as_connection, key_input)

        await asyncio.gather(async_io(key))

    @pytest.mark.asyncio
    async def test_neg_get_with_non_existent_namespace(self):
        """
        Invoke get() for non-existent namespace.
        """
        key = ("namespace", "demo", 1)

        async def async_io(key_input=None, policy_input=None):
            with pytest.raises(e.ClientError):
                await io.get(self.as_connection, key_input)

        await asyncio.gather(async_io(key))

    @pytest.mark.asyncio
    @pytest.mark.parametrize(
        "_input, _expected",
        [
            (("test", "demo", 1), {"name": "john", "age": 1}),
        ],
    )
    async def test_neg_get_remove_key_and_check_get(self, _input, _expected, put_data):
        """
        Invoke get() for a record having string data.
        """
        put_data(self.as_connection, _input, _expected)
        self.as_connection.remove(_input)

        async def async_io(key_input=None, policy_input=None):
            try:
                rec = await io.get(self.as_connection, key_input)
                assert rec is None
            except e.RecordNotFound as exception:
                assert exception.code == 2

        await asyncio.gather(async_io(_input))

    @pytest.mark.asyncio
    async def test_neg_get_with_only_key_no_connection(self):
        """
        Invoke get() with a key and not policy's dict no connection
        """
        key = ("test", "demo", 1)
        config = TestBaseClass.get_connection_config()
        client1 = aerospike.client(config)
        client1.close()

        async def async_io(key_input=None, policy_input=None):
            with pytest.raises(e.ClusterError):
                key, _, _ = await io.get(client1, key_input)

        await asyncio.gather(async_io(key))



import aerospike
import sys
from aerospike import exception as ex

config = {'hosts': [('127.0.0.1', 3000)]}
client = aerospike.client(config).connect()

# Create a client and connect it to the cluster
try:
    client = aerospike.client(config).connect()
    client.truncate('test', "demo", 0)
except ex.ClientError as e:
    print("Error: {0} [{1}]".format(e.msg, e.code))
    sys.exit(1)

# Remove old indices
try:
    client.index_remove("test", "scoreIndex")
    client.index_remove("test", "eloIndex")
except ex.AerospikeError as e:
    # Ignore if no indices found
    pass

# Insert 4 records
keyTuples = [("test", "demo", f"player{i}") for i in range(4)]
bins = [
    {"score": 100, "elo": 1400},
    {"score": 20, "elo": 1500},
    {"score": 10, "elo": 1100},
    {"score": 200, "elo": 900}
]
for keyTuple, bin in zip(keyTuples, bins):
    client.put(keyTuple, bin)

query = client.query('test', 'demo')

# Queries require a secondary index for each bin name
client.index_integer_create("test", "demo", "score", "scoreIndex")
client.index_integer_create("test", "demo", "elo", "eloIndex")



# -*- coding: utf-8 -*-
##########################################################################
# Copyright 2013-2021 Aerospike, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##########################################################################


from __future__ import print_function

import aerospike
import sys

from optparse import OptionParser
from aerospike import exception as e

##########################################################################
# Options Parsing
##########################################################################

usage = "usage: %prog [options]"

optparser = OptionParser(usage=usage, add_help_option=False)

optparser.add_option(
    "--help", dest="help", action="store_true",
    help="Displays this message.")

optparser.add_option(
    "-U", "--username", dest="username", type="string", metavar="<USERNAME>",
    help="Username to connect to database.")

optparser.add_option(
    "-P", "--password", dest="password", type="string", metavar="<PASSWORD>",
    help="Password to connect to database.")

optparser.add_option(
    "-h", "--host", dest="host", type="string", default="127.0.0.1", metavar="<ADDRESS>",
    help="Address of Aerospike server.")

optparser.add_option(
    "-p", "--port", dest="port", type="int", default=3000, metavar="<PORT>",
    help="Port of the Aerospike server.")

optparser.add_option(
    "--timeout", dest="timeout", type="int", default=1000, metavar="<MS>",
    help="Client timeout")

optparser.add_option(
    "--read-timeout", dest="read_timeout", type="int", default=1000, metavar="<MS>",
    help="Client read timeout")

optparser.add_option(
    "-n", "--namespace", dest="namespace", type="string", default="test", metavar="<NS>",
    help="Port of the Aerospike server.")

optparser.add_option(
    "-s", "--set", dest="set", type="string", default="demo", metavar="<SET>",
    help="Port of the Aerospike server.")

(options, args) = optparser.parse_args()

if options.help:
    optparser.print_help()
    print()
    sys.exit(1)

##########################################################################
# Application
##########################################################################

try:

    # ----------------------------------------------------------------------------
    # Connect to Cluster
    # ----------------------------------------------------------------------------

    config = {
        'hosts': [(options.host, options.port)],
        'policies': {
            'total_timeout': options.timeout
        }
    }
    client = aerospike.client(config).connect(
        options.username, options.password)

    # ----------------------------------------------------------------------------
    # Perform Operation
    # ----------------------------------------------------------------------------

    try:
        namespace = options.namespace if options.namespace and options.namespace != 'None' else None
        set = options.set if options.set and options.set != 'None' else None
        smile = u"smilé"

        key = (namespace, set, smile)
        bins = {'smiley': smile, 'smile_count': 1, 'mood': 'happy'}
        print("Storing ", bins, "at a record identified by the tuple", key)
        # overwrite the record if it exists, otherwise create it
        client.put(key, bins,
                   policy={'exists': aerospike.POLICY_EXISTS_CREATE_OR_REPLACE,
                           'key': aerospike.POLICY_KEY_SEND})
        print("Retrieving the record from the server for comparison")
        (key, meta, record) = client.get(
            key, policy={'total_timeout': options.read_timeout})
        print("The value of the 'smiley' bin is", record['smiley'], "\n")
        print("By the way, this record has been written", meta['gen'], "times")
        future_gen = str(int(meta['gen']) + 2)
        print("Expect the record generation to be",
              future_gen, "with two more write operations\n")

        # add a dictionary under a bin named 'data'
        bins = {'data': {'smiley_key': smile, smile: 'this is a smiley '}}
        print("Storing ", bins, "at the record", key)
        client.put(key,  bins)
        (key, metadata, bins) = client.get(key)
        print("The value of the 'smiley_key' of the 'data' bin is",
              bins['data']['smiley_key'], "\n")
        # print("The value of the", smile, " key is:",
        #      bins['data'][smile], "\n")

        # append to the value of the smile key
        print("Before appending, the value of the 'mood' key is:",
              bins['mood'])
        client.append(key, 'mood', smile)
        (key, metadata, bins) = client.get(key)
        print("After appending, the value of the 'mood' key is:",
              bins['mood'], "\n")

        # prepend to the value of the smile key
        print("Before prepending, the value of the 'mood' key is:",
              bins['mood'])
        client.prepend(key, 'mood', smile)
        (key, metadata, bins) = client.get(key)
        print("After prepending, the value of the 'mood' key is:",
              bins['mood'], "\n")

        # multiple operations on the record using the operate() method
        ops = [{'bin': 'smiley', 'op': aerospike.OPERATOR_APPEND, 'val': smile},
               {'bin': 'smile_count', 'op': aerospike.OPERATOR_INCR, 'val': 5},
               {'bin': 'smiley', 'op': aerospike.OPERATOR_READ}]
        print("Setting the following multiops on the same record\n", ops)
        (key, meta, bins) = client.operate(key, ops)
        print("The value of the 'smiley' bin is", bins['smiley'], "\n")

        print("Displaying the key, metadata, and bins of the record")
        (key, meta, bins) = client.get(key)
        print(key)
        print(meta)
        print(bins, "\n")
        client.remove(key)

        # example of a bytearray primary key
        print("Save a new record with a bytearray primary key")
        smiley_pk = smile.encode("utf-8")
        client.put((namespace, set, smiley_pk), {'smiley': smile, 'smiley_pk':
                                                 smiley_pk})
        print("Display the bins of a record with a bytearray key")
        (key, meta, bins) = client.get((namespace, set, smiley_pk))
        print(bins)
        print("The value of the 'smiley_pk' bin is", bins['smiley_pk'], "\n")
        client.remove(key)
        exitCode = 0
    except Exception as exception:
        print("error: {0}".format(exception), file=sys.stderr)
        exitCode = 2

    # ----------------------------------------------------------------------------
    # Close Connection to Cluster
    # ----------------------------------------------------------------------------

    client.close()

except e.ClientError as exception:
    print("Error: {0} [{1}]".format(exception.msg, exception.code))
    exitCode = 3

##########################################################################
# Exit
##########################################################################

sys.exit(exitCode)

#



# -*- coding: utf-8 -*-
##########################################################################
# Copyright 2013-2021 Aerospike, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##########################################################################

from __future__ import print_function

import aerospike
import json
import re
import sys
import os.path

from optparse import OptionParser
from aerospike import predicates as p

##########################################################################
# Option Parsing
##########################################################################

usage = "usage: %prog [options] [where]"


def query_callback(option, opt, value, parser):
    setattr(parser.values, option.dest, value.split(','))

optparser = OptionParser(usage=usage, add_help_option=False)

optparser.add_option(
    "-h", "--host", dest="host", type="string", default="127.0.0.1", metavar="<ADDRESS>",
    help="Address of Aerospike server.")

optparser.add_option(
    "-p", "--port", dest="port", type="int", default=3000, metavar="<PORT>",
    help="Port of the Aerospike server.")

optparser.add_option(
    "--help", dest="help", action="store_true",
    help="Displays this message.")

optparser.add_option(
    "-n", "--namespace", dest="namespace", type="string", default="test", metavar="<NS>",
    help="Port of the Aerospike server.")

optparser.add_option(
    "-s", "--set", dest="set", type="string", default="demo", metavar="<SET>",
    help="Port of the Aerospike server.")

optparser.add_option(
    "-m", "--module", dest="module", type="string",
    help="UDF Module.")

optparser.add_option(
    "-f", "--function", dest="function", type="string",
    help="UDF Function.")

optparser.add_option(
    "-a", "--arg", dest="arguments", type="string", action="callback",
    callback=query_callback,  help="UDF Arguments.")

optparser.add_option(
    "-b", "--bins", dest="bins", type="string", action="append",
    help="Bins to select from each record.")

optparser.add_option(
    "--show-key", dest="show_key", action="store_true",
    help="If set, displays the key/digest.")

optparser.add_option(
    "--show-meta", dest="show_meta", action="store_true",
    help="If set, displays the metadata.")


(options, args) = optparser.parse_args()

if options.help:
    optparser.print_help()
    print()
    sys.exit(1)

if len(args) > 1:
    optparser.print_help()
    print()
    sys.exit(1)

##########################################################################
# Client Configuration
##########################################################################

config = {
    'hosts': [(options.host, options.port)],
    'lua': {
        'user_path': os.path.dirname(__file__)
    }
}

##########################################################################
# Application
##########################################################################

exitCode = 0

try:

    # ----------------------------------------------------------------------------
    # Connect to Cluster
    # ----------------------------------------------------------------------------

    client = aerospike.client(config).connect()

    # ----------------------------------------------------------------------------
    # Perform Operation
    # ----------------------------------------------------------------------------

    try:

        query_id = 0
        re_bin = "(.{1,14})"
        re_str_eq = "\s+=\s*(?:(?:\"(.*)\")|(?:\'(.*)\'))"
        re_int_eq = "\s+=\s*(\d+)"
        re_int_rg = "\s+between\s+\(\s*(\d+)\s*,\s*(\d+)\s*\)"
        re_w = re.compile("%s(?:%s|%s|%s)" %
                          (re_bin, re_str_eq, re_int_eq, re_int_rg))

        namespace = options.namespace if options.namespace and options.namespace != 'None' else None
        set = options.set if options.set and options.set != 'None' else None

        q = None

        for i, param in enumerate(options.arguments):
            if param.isdigit():
                options.arguments[i] = int(param)

        if len(args) == 1:
            w = re_w.match(args[0])
            if w is not None:

                # If predicate is provided, then perform a query

                if w.group(2):
                    b = w.group(1)
                    v = w.group(2)
                    query_id = client.query_apply(options.namespace,
                                                  options.set, p.equals(
                                                      b, v), options.module,
                                                  options.function, options.arguments)
                elif w.group(3):
                    b = w.group(1)
                    v = w.group(3)
                    query_id = client.query_apply(options.namespace,
                                                  options.set, p.equals(
                                                      b, v), options.module,
                                                  options.function, options.arguments)
                elif w.group(4):
                    b = w.group(1)
                    v = int(w.group(4))
                    query_id = client.query_apply(options.namespace,
                                                  options.set, p.equals(
                                                      b, v), options.module,
                                                  options.function, options.arguments)
                elif w.group(5) and w.group(6):
                    b = w.group(1)
                    l = int(w.group(5))
                    u = int(w.group(6))
                    query_id = client.query_apply(options.namespace,
                                                  options.set, p.between(
                                                      b, l, u), options.module,
                                                  options.function, options.arguments)

        while True:
            response = client.job_info(query_id, aerospike.JOB_QUERY)
            if response['status'] == aerospike.JOB_STATUS_COMPLETED:
                break

        if response['status'] == aerospike.JOB_STATUS_COMPLETED:
            print("Background query is successful")
        else:
            print("Query_apply failed")

    except Exception as eargs:
        print("error: {0}".format(eargs), file=sys.stderr)
        exitCode = 2

    # ----------------------------------------------------------------------------
    # Close Connection to Cluster
    # ----------------------------------------------------------------------------

    client.close()

except Exception as eargs:
    print("error: {0}".format(eargs), file=sys.stderr)
    exitCode = 3


##########################################################################
# Exit
##########################################################################

sys.exit(exitCode)



from aerospike import exception as e
import time


def index_found_in_info_res(res, index_name):
    found = False
    for key in res:
        value = res[key]
        response = value[1]
        if response:
            if index_name in response:
                return True
    return found


def ensure_dropped_index(client, namespace, index_name):
    try:
        client.index_remove(namespace, index_name)
    except (e.IndexNotFound, e.InvalidRequest):
        pass
    retries = 0
    while retries < 10:
        responses = client.info_all("sindex")
        if not index_found_in_info_res(responses, index_name):
            return
        time.sleep(0.5)
        retries += 1
